"""Utils about encryption
"""

import hashlib


def md5(text):
    """Calculate MD5 hash of text
    """
    if isinstance(text, str):
        text = text.encode('utf-8')
    return hashlib.md5(text).hexdigest()


"""Utils enhancing inspect module"""

__all__ = [
    'is_instance_method_by_signature',
    'has_kwarg_by_sig',
    'args_to_kwargs_by_sig',
    'get_param_types',
]

import inspect
import typing


def is_instance_method_by_signature(
    func: typing.Callable
) -> bool:
    """Check if the function is an instance method by checking its signature.

    If the function has a 'self' parameter and positioned as the first parameter,
    it is considered an instance method.
    """

    sig = inspect.signature(func)
    params = sig.parameters
    if len(params) > 0:
        first_param = tuple(params.values())[0]
        if first_param.name == 'self':
            return True
    return False


def has_kwarg_by_sig(
    func: typing.Callable, kwarg_name: str
) -> bool:
    """Check if the function has a specific keyword argument by checking its signature.

    :param func: The function to check
    :param kwarg_name: The name of the keyword argument to check for
    """
    sig = inspect.signature(func)
    params = sig.parameters
    return any(
        param.name == kwarg_name and param.kind in (
            inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_KEYWORD
        )
        for param in params.values()
    )


def args_to_kwargs_by_sig(
    func: typing.Callable, *args,
    offset: int = 0,
    try_default: bool = True,
) -> typing.Dict[str, typing.Any]:
    """Convert positional arguments to keyword arguments based on the function's signature.

    :param func: The function to check
    :param args: The positional arguments to convert
    :param offset: The number of positional arguments to skip (on sig arguments)
    :param try_default: If True, use default values for missing positional arguments
    """
    sig = inspect.signature(func)
    params = sig.parameters
    kwargs = {}

    for i, (name, param) in enumerate(params.items()):
        if i < offset:
            continue
        if i < len(args):
            kwargs[name] = args[i-offset]
        else:
            if try_default:
                if param.default is not param.empty:
                    kwargs[name] = param.default

    return kwargs

def get_param_types(
    func: typing.Callable
) -> typing.Mapping[str, type]:
    """Get the parameter types of a function based on its signature.

    :param func: The function to inspect
    :return: A dictionary mapping parameter names to their types
    """
    sig = inspect.signature(func)
    return {
        name: (
            param.annotation
            if param.annotation is not inspect.Parameter.empty
            else typing.Any
        )
        for name, param in sig.parameters.items()
    }




"""Task registry module.
"""

__all__ = [
    'TaskRegistry',
    'TaskEntry',
    'listen_to'
]

import asyncio
import copy
import inspect
import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit

from ..exceptions import TaskHandlerNotFound
from .._types import PathParamsT, CallableTV
from ..transport.base import BaseTransporter
from .result import Body, JsonBody
from ..task.context import BaseTaskContext
from ..core.middleware import BaseMiddleware
from .main import TaskID, Method
from . import TaskHandler
from ..utils.inspect_ import get_param_types

if typing.TYPE_CHECKING:
    from ..manager import BaseManager


class TaskEntry(BaseMiddleware):
    """BlueFirmament TaskEntry

    A mapping from TaskID to a couple of TaskHandler(s).

    Work as a middleware, run this middleware will concurrently
    run all handlers in this entry.

    Can be stored as key in dict or element in set.

    :ivar path_params: Path parameters resolved from the looked up TaskID.
        Will be set on the copy of this entry.
    """

    def __init__(
        self,
        task_id: TaskID,
        *handlers: TaskHandler | typing.Callable
    ) -> None:
        self.__task_id = task_id
        self.__handlers: typing.List[TaskHandler] = list(
            handler if isinstance(handler, TaskHandler) else TaskHandler(function=handler)
            for handler in handlers
        )
        self.path_params: dict[str, PathParamsT] = {}

    @property
    def id(self):
        return self.__task_id

    @property
    def handlers(self):
        return self.__handlers

    def fork(
        self,
        path_prefix: str = ""
    ) -> typing.Self:
        """Fork this entry with a new TaskID

        :param path_prefix:
            Prefix task_id path with this.
        :return: Forked TaskEntry
        """
        return TaskEntry(
            self.__task_id.fork(
                path_prefix=path_prefix
            ),
            *self.__handlers
        )

    def is_dynamic(self) -> bool:
        return self.__task_id.is_dynamic()

    def is_match(self, task_id: TaskID) -> Opt[dict]:
        """Whether a task_id match this entry's task_id
        """
        return self.__task_id.is_match(task_id)

    def add_handler(self, task_handler: TaskHandler):
        """Add a TaskHandler
        """
        self.__handlers.append(task_handler)

    def set_manager_cls(self, manager_cls: typing.Type["BaseManager"]):
        """Set manager class for all handlers in this entry.
        """
        for handler in self.__handlers:
            handler.set_manager_cls(manager_cls)

    async def __call__(self, *, next_, task_context: 'BaseTaskContext'):
        """Run all handlers in this entry concurrently.

        If multiple results, set task result body to a JsonBody which is a list
        wrapped those results.
        If only one result (where when one handler), set task result body to it.
        """
        coros: list[typing.Awaitable[Body]] = []
        for handler in self.__handlers:
            coros.append(handler(
                task_context=task_context, path_params=self.path_params
            ))

        results = await asyncio.gather(*coros)

        if len(results) == 1:
            task_context._task_result.body = results[0]
        else:
            task_context._task_result.body = JsonBody([
                result
                for result in results
            ])

        await next_()


class TaskRegistry:
    """BlueFirmament Task Registry

    A bunch of task entries that you can look up by TaskID.
    """

    def __init__(self, 
        name: str = 'router',
        path_prefix: str = ''
    ):
        """
        :param path_prefix: 
            Prefix added to every record path
            registered to this router.

            Can be ``/abc/{var}`` or ``abc/{var}``, but don't
            end with a slash.
        """
        self.__static_entries: dict[TaskID, TaskEntry] = dict()
        self.__dynamic_entries: list[TaskEntry] = list()
        self.__path_prefix = path_prefix
        self.__name = name

    @property
    def name(self): return self.__name
    @property
    def static_entries(self): return self.__static_entries
    @property
    def dynamic_entries(self): return self.__dynamic_entries

    def add_entry(self, entry: TaskEntry):
        entry = entry.fork(path_prefix=self.__path_prefix)
        if not entry.is_dynamic():
            self.__static_entries[entry.id] = entry
        else:
            self.__dynamic_entries.append(entry)

    def add_entries(self, entries: typing.Iterable[TaskEntry]):
        for entry in entries:
            self.add_entry(entry)
        
    def add_handler(
        self,
        method: Opt[Method] = None,
        path: Opt[str] = None,
        task_id: Opt[TaskID] = None,
        handler: Opt[TaskHandler] = None,
        function: Opt[typing.Callable] = None,
        handler_manager_cls: Opt[typing.Type["BaseManager"]] = None,
    ):
        """Bind a task handler to Task(ID).

        :param method:
        :param path: 
            Must start with a slash and ends with no slash.
        :param task_id:
            Replace method and path and path_prefix won't be applied.
        :param handler: A task handler.
        """
        if handler is None:
            if not (function is None or handler_manager_cls is None):
                handler = TaskHandler(
                    function=function,
                    manager_cls=handler_manager_cls
                )
            else:
                raise ValueError("provide handler or inner_handler and handler_manager")
        
        if task_id is None:
            if not (method is None or path is None):
                task_id = TaskID(
                    method, self.__path_prefix + path,
               )
            else:
                raise ValueError("provide task_id or method and path")

        try:
            if not task_id.is_dynamic():
                entry = self.lookup(task_id)
            else:
                # lookup in dynamic entries
                entry = next(
                    (e for e in self.__dynamic_entries if e.is_match(task_id)),
                    None
                )
            if entry is None:
                raise KeyError
        except KeyError:
            # no such entry, create one
            entry = TaskEntry(task_id, handler)

        if not task_id.is_dynamic():
            self.__static_entries[entry.id] = entry
        else:
            self.__dynamic_entries.append(entry)

    def merge(self, to_merge: "TaskRegistry"):
        """Merge another task registry's entries.

        Every entry to be merged will be prefixed with the path_prefix.
        (Of course on the forked entry)
        """
        for entry in to_merge.static_entries.values():
            entry = entry.fork(self.__path_prefix)
            self.__static_entries[entry.id] = entry
        for entry in to_merge.dynamic_entries:
            entry = entry.fork(self.__path_prefix)
            self.__dynamic_entries.append(entry)

    def lookup(
        self,
        task_id: TaskID,
    ) -> TaskEntry:
        """Lookup a task entry by task_id.

        :param task_id: The task ID to lookup, must be static.
        :raise NotFound: If no task entry matched.
        :raise TypeError: If task_id is dynamic.
        :returns:
            The (shallow) copy of the matched TaskEntry
            with path_params set.
        """
        if task_id.is_dynamic(allow_method_dynamic=True):
            raise TypeError("Cannot lookup a dynamic task_id")

        entry = self.__static_entries.get(task_id, None)
        if entry is not None:
            return entry
        else:
            # lookup in dynamic entries
            for entry in self.__dynamic_entries:
                match_res = entry.is_match(task_id)
                if match_res is not None:
                    entry = copy.copy(entry)
                    entry.path_params = match_res
                    return entry
                else:
                    continue

        raise TaskHandlerNotFound(
            task_id=task_id,
        )


def listen_to(
    method: Opt[Method | str],
    path: str,
    separator: str = "/",
    transporters: Opt[typing.Iterable[str | BaseTransporter]] = None,
):
    """Make the function a handler to a task.

    :param transporters:
        Only tasks from these transporters will be handled by this handler.
        None for default transporter (you must have a transporter named "default").

    Will wrap decorated function to a TaskEntry.
    With support of :meth:`blue_firmament.manager.ManagerMetaclass`,
    this entry will be added to manager task registry.
    Finally, with :meth:`blue_firmament.task.TaskRegistry.merge` or
    :meth:`blue_firmament.core.BlueFirmamentApp.add_manager`,
    manager task registry's entries will be merged into application
    task registry.
    """
    def wrapper(handler: CallableTV) -> CallableTV:
        return typing.cast(CallableTV, (
            tuple(transporters or ("default",)),
            TaskEntry(
                TaskID(
                    method=method, path=path,
                    separator=separator,
                    param_types=get_param_types(handler),
                ),
                handler
            )
        )) # lie to type checker
    return wrapper



"""Task module of BlueFirmament.

Task is the service unit of an BlueFirmament Application.

Design doc: :doc:`design/task`
"""

__all__ = [
    'TaskID',
    'Task',
    'TaskStatus',
    'TaskMetadata',
    'TaskResult',
    'TaskHandler',
    'TaskRegistry',
    'TaskEntry',
    'listen_to',
    'Method',
    'LazyParameter'
]

from .main import (
    TaskID, Task, TaskMetadata, Method, LazyParameter
)
from .result import (
    TaskStatus, TaskResult
)
from .handler import TaskHandler
from .registry import (
    TaskRegistry, TaskEntry, listen_to
)



"""BlueFirmament Task Context Module.
"""

__all__ = [
    "BaseTaskContext",
    "SoBaseTC",
    "ExtendedTaskContext",
    "CommonTaskContext",
    "SoCommonTC"
]


from .main import (
    BaseTaskContext, SoBaseTC, ExtendedTaskContext
)
from .common import (
    CommonTaskContext, SoCommonTC
)



"""Blue Firmament - A Python backend framework.
"""

__version__ = "0.1.2"
__name__ = "blue_firmament"
__all__ = [
    "listen_to",
    "Method",
    "BlueFirmamentApp"
]


from .core import BlueFirmamentApp
from .task import listen_to, Method

from .utils import json_
json_.override_json_encoder(json_.JsonEncoder)


"""Task result module.
"""

__all__ = [
    "Body",
    "EmptyBody",
    "JsonBody",
    "StreamingBody",
    "TaskStatus",
    "TaskResult"
]

import abc
import json
import typing
from typing import Annotated as Anno, Optional as Opt, Literal as Lit
import enum

from . import TaskMetadata
from ..scheme import BaseScheme
from ..utils.main import singleton
from ..utils.typing_ import JsonDumpable


TV = typing.TypeVar('TV')
class Body(abc.ABC, typing.Generic[TV]):
    """Task Result Body
    """

    def __init__(self, data: TV = None) -> None:
        self._data: TV = data

    def dump_to_bytes(self, encoding: str = "utf-8") -> bytes:
        """Dump body to bytes.

        Extend supported value types by overriding this method,
        currently supported types are:
        - bytes
        - str (will be encoded to bytes using specified encoding)
        """
        if isinstance(self._data, bytes):
            return self._data
        if isinstance(self._data, str):
            return self._data.encode(encoding)
        raise NotImplementedError("This body is not supporting bytes serialization")

    def dump_to_dict(self) -> dict:
        if isinstance(self._data, dict):
            return self._data
        raise NotImplementedError("This body is not supporting dict serialization")

    def dump_to_json(self) -> str:
        try:
            return json.dumps(self._data, ensure_ascii=False)
        except TypeError:
            raise NotImplementedError('This body is not supporting JSON serialization')

    def dump_to_str(self) -> str:
        if isinstance(self._data, str):
            return self._data
        raise NotImplementedError("This body is not supporting str serialization")

    async def cleanup(self) -> None:
        pass


@singleton
class EmptyBody(Body):
    """Nothing here
    """

    def dump_to_dict(self) -> dict:
        return {}

    def dump_to_bytes(self, encoding: str = "utf-8") -> bytes:
        return b''

    def dump_to_json(self) -> str:
        return ''

class PlainTextBody(Body[str]):
    """Plain text body
    """

    def dump_to_str(self) -> str:
        return self._data

class JsonBody(Body[JsonDumpable]):

    def __init__(self, data: JsonDumpable) -> None:
        super().__init__(data)

    def dump_to_dict(self) -> dict:
        if not isinstance(self._data, (dict, BaseScheme)):
            raise TypeError(f'cannot dump {type(self._data)} to dict')
        if isinstance(self._data, BaseScheme):
            return self._data.dump_to_dict()
        return self._data

    def dump_to_bytes(self, encoding: str = "utf-8") -> bytes:
        return self.dump_to_json().encode(encoding)

    def dump_to_json(self) -> str:
        from ..utils.json_ import dumps_to_json
        return dumps_to_json(self._data)


class StreamingBody(Body):
    """An iterable body that yield events.

    :ivar __generator: 事件生成器
    :ivar __cleanup: 清理函数
    """

    type GeneratorType = typing.AsyncGenerator["Body", None]

    def __init__(
        self,
        generator: GeneratorType,
        cleanup: typing.Callable
    ) -> None:
        """
        :param generator:
            Event generator, yields Body other than StreamingBody.
        :param cleanup: Cleanup function to call when sending stopped unexpectedly.
        """
        super().__init__()
        self.__generator = generator
        self.__cleanup = cleanup

    def dump_to_bytes(self, encoding: str = "utf-8") -> bytes:
        raise NotImplementedError('This body is not supporting bytes serialization')

    def __aiter__(self) -> GeneratorType:
        return self.__generator
    
    async def cleanup(self) -> None:
        self.__cleanup()


class TaskStatus(enum.Enum):
    OK = 200
    CREATED = 201
    DELETED = 204
    BAD_REQUEST = 400
    UNPROCESSABLE_ENTITY = 422
    UNAUTHORIZED = 401
    FORBIDDEN = 403
    NOT_FOUND = 404
    CONFLICT = 409
    UNAVAILABLE_FOR_LEGAL_REASONS = 451
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    SERVICE_UNAVAILABLE = 503


class TaskResult:

    def __init__(self,
        status: TaskStatus = TaskStatus.OK,
        body: Body = EmptyBody(),
        metadata: Opt[TaskMetadata] = None,
    ):
        self.__body: Body = body
        self.__status: TaskStatus = status
        self.__metadata: TaskMetadata = metadata or TaskMetadata()

    @property
    def metadata(self) -> TaskMetadata:
        return self.__metadata

    @property
    def body(self) -> Body:
        return self.__body

    @body.setter
    def body(self, value: Body) -> None:
        self.__body = value

    @property
    def status(self) -> TaskStatus:
        return self.__status
    
    @status.setter
    def status(self, value: TaskStatus) -> None:
        self.__status = value



"""Utils related to Enums."""

__all__ = [
    'dump_enum',
    'load_enum',
]

import typing
import enum


DumpEnumValueType = typing.TypeVar('DumpEnumValueType')
@typing.overload
def dump_enum(enum_member: enum.Enum) -> DumpEnumValueType:
    ...
@typing.overload
def dump_enum(enum_member: DumpEnumValueType) -> DumpEnumValueType:
    ...
@typing.overload
def dump_enum(enum_member: None) -> None:
    ...
def dump_enum(enum_member: enum.Enum | DumpEnumValueType | None) -> DumpEnumValueType | None:
    """Dump enum member to its value.
    """
    if enum_member is None:
        return None
    if isinstance(enum_member, enum.Enum):
        return enum_member.value
    return enum_member

def dump_enums(enums: typing.Iterable[enum.Enum]) -> typing.Iterable[DumpEnumValueType]:
    """Dump iterable of enum members to their values.

    :param enums: Iterable of enum members.
    :returns: Iterable of enum member values.
    """
    return (dump_enum(enum_member) for enum_member in enums)

EnumType = typing.TypeVar('EnumType', bound=enum.Enum)
@typing.overload
def load_enum(
    enum_class: typing.Type[EnumType],
    name: str | int | EnumType,
) -> EnumType:
    ...
@typing.overload
def load_enum(
    enum_class: typing.Type[EnumType],
    name: None,
) -> None:
    ...
def load_enum(
    enum_class: typing.Type[EnumType],
    value: str | int | EnumType | None,
) -> EnumType | None:
    """Resolve enum member from value.

    :param enum_class: The enum class.
    :param value: The enum member value or a member of the enum class.

    :raise ValueError: If no enum member for this value in the enum class.
    :returns: Enum member. If you pass None as name, returns None.
    """
    if value is None:
        return None
    if isinstance(value, enum.Enum):
        return value
    else:
        try:
            return enum_class(value)
        except ValueError as e:
            raise ValueError(
                f"Value '{value}' is not a valid member of enum '{enum_class.__name__}'"
            ) from e



"""BlueFirmament TopLevel Exceptions

TODO exception factory to manage logger automatically
"""

__all__ = [
    "BlueFirmamentException",
    "BFExceptionTV",
    "InternalError",
    "ClientError",
    "RequestFailed",
    "ExternalError",
    "Retryable",
    "MaxRetriesExceeded",
    "ParamsInvalid",
    "AtLeastOne",
    "NotImplemented",
    "NotFound",
    "TaskHandlerNotFound",
    "DuplicateOrConflict",
    "Duplicate",
    "Conflict",
    "InvalidStatusTransition",
    "Unauthorized",
    "Forbidden",
    "TooManyRequests",
]

import enum
import abc
import typing
from typing import Optional as Opt
from .utils.typing_ import JsonDumpable
from .task import TaskStatus, TaskID
from .log.main import get_logger

LOGGER = get_logger(__name__)
"""默认日志记录器
"""

if typing.TYPE_CHECKING:
    from .scheme import BaseScheme


BFExceptionTV = typing.TypeVar("BFExceptionTV", bound="BlueFirmamentException")
"""碧霄异常类型变量"""


class BlueFirmamentException(Exception, abc.ABC):
    """碧霄异常基类

    :ivar errmsg: 错误描述
    :ivar

    """

    def __init__(self, errmsg: str | dict = "Exception occured", *args, **kwargs):
        if isinstance(errmsg, str):
            errmsg = {"errmsg": errmsg, **kwargs}

        self.errmsg: dict = errmsg

        super().__init__(errmsg, *args)

    def apply_to_task_result(self, response) -> None:
        """将异常序列化为响应体

        直接修改已有的响应对象的数据

        :param response: 响应对象
        """

    def dump_to_jsonable(self) -> JsonDumpable:
        """序列化为可JSON序列化的类型"""
        return self.errmsg

    @property
    @abc.abstractmethod
    def task_status(self) -> TaskStatus: ...

    @classmethod
    def from_python_exception(cls, exception: Exception) -> "BlueFirmamentException":
        """从Python异常创建碧霄异常

        :param exception: Python异常
        """
        if isinstance(exception, ValueError):
            return ParamsInvalid()

        return InternalError()

    @property
    def name(self) -> str:
        """异常名称"""
        return self.__class__.__name__

    def dump_details_to_str(self) -> str:
        """将详细信息序列化为字符串"""
        return "\n".join([f"{k}: {v}" for k, v in self.errmsg.items()])

    def dump_details_to_dict(self) -> dict:
        """将详细信息序列化为字典"""
        return self.errmsg

    def __str__(self) -> str:
        return f"{self.name}: \n{self.dump_details_to_str()}"


class InternalError(BlueFirmamentException):
    """Exception raised cause of server's fault."""

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.INTERNAL_SERVER_ERROR


class ClientError(BlueFirmamentException):
    """Exception raised cause of client's fault"""

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.BAD_REQUEST


class RequestFailed(InternalError):
    """请求失败

    请求没有到达目标服务
    """

    def __init__(self, response, *args, **kwargs):
        super().__init__({"response": response}, *args, **kwargs)


class ExternalError(BlueFirmamentException):
    """外部错误

    任何外部应用程序与服务导致的异常都将转换为该异常
    """

    def __init__(
        self,
        service_name: str,
        logger=LOGGER,
        **errinfo: typing.Any,
    ):
        """
        :param service_name: 外部服务名称
        :param errcode: 错误码

            可以是错误码、错误信息等
        """

        super().__init__({"service_name": service_name, **errinfo}, logger)

    def __getitem__(self, key: str) -> typing.Any:
        """获取错误信息"""
        return self.errmsg[key]

    @property
    def task_status(self):
        return TaskStatus.SERVICE_UNAVAILABLE


class Retryable(BlueFirmamentException):
    """The exception can be eliminated by retrying the operation."""

    def __init__(self, delay: Opt[float] = None):
        """
        :param delay: retry after this many seconds
        """
        super().__init__("please retry after %s seconds" % delay)
        self.__delay = delay

    @property
    def delay(self) -> Opt[float]:
        return self.__delay

    @property
    def task_status(self):
        return TaskStatus.SERVICE_UNAVAILABLE


class MaxRetriesExceeded(BlueFirmamentException):
    def __init__(self):
        super().__init__("max retries exceeded")

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.SERVICE_UNAVAILABLE


class ParamsInvalid(ClientError, ValueError):
    """参数无效

    场景
    ------
    - 类型不合法
    - 值不合法
    - 缺少参数
    - 通过参数获得（计算、数据库访问）的值不合法
    """

    def __init__(self, msg: str = "", **params):
        super().__init__(
            {"message": "Invalid parameter(s), %s" % msg, "parameters": params}
        )

    @property
    def task_status(self):
        return TaskStatus.UNPROCESSABLE_ENTITY


class ParamRequired(ClientError, ValueError):
    """Exception for not giving required parameter."""

    def __init__(self, param_name: str, **kwargs):
        super().__init__("parameter required", param=param_name)


class AtLeastOne(ParamsInvalid):
    """至少需要一个参数

    在这些参数中，至少提供一个（有效）
    """

    def __init__(
        self,
        *params: typing.Any,
    ):
        """
        :param params: 参数名称
        """
        self.__params = params

    def __str__(self) -> str:
        return (
            f"{self.name}: at least one of following parameters should be provided:\n"
            + "\n    ".join(self.__params)
        )


class NotImplemented(InternalError, NotImplementedError):
    def __init__(self, service_name: str, *args, **kwargs):
        """
        :param service_name: 未实现的服务名
        """
        errmsg = "service %s is not implemented" % service_name
        super().__init__(errmsg, *args, **kwargs)

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.NOT_IMPLEMENTED


class NotFound(ClientError, KeyError):
    """未找到"""

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.NOT_FOUND


class TaskHandlerNotFound(NotFound):
    """没有该任务的处理器"""

    def __init__(self, task_id: TaskID):
        super().__init__("task handler not found", task_id=task_id)

    @property
    def task_id(self):
        return self.errmsg.get("task_id", None)


class DuplicateOrConflict(InternalError):
    """冲突或重复"""

    def __init__(
        self,
        *resource: tuple[typing.Any, typing.Any],
        operation: str,
        msg: str = "",
    ):
        """
        :param resource: 资源的的元组
            用（资源名，资源标识符）表示一个资源实例
        :param operation: 正在对资源进行的操作
        :param errmsg: 附加描述
        """

        super().__init__(
            f"Cannot perform {operation} on {resource[0]}: {resource[1]} cause: \n{msg}"
        )

    @classmethod
    def from_scheme(
        cls,
        scheme: "BaseScheme",
        operation: str,
        errmsg: str = "",
    ) -> "DuplicateOrConflict":
        """从数据模型 实例创建冲突异常"""
        return cls(
            (scheme.dal_path(), scheme[scheme._get_key_field()]),
            operation=operation,
            msg=errmsg,
        )

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.CONFLICT


class Duplicate(DuplicateOrConflict):
    def __init__(self, msg: str):
        self.__msg = msg

    def dump_details_to_str(self) -> str:
        return self.__msg


class Conflict(DuplicateOrConflict):
    def __init__(self, msg: str, **kwargs):
        self.__msg = msg
        self.__context = kwargs

    def dump_details_to_str(self) -> str:
        return self.__msg


class InvalidStatusTransition(Conflict):
    """无效的状态转换

    不可以从当前状态切换到目标状态
    """

    def __init__(self, status_name: str, current_status, target_status):
        self.__status_name = status_name
        self.__current_status = current_status
        self.__target_status = target_status

    @classmethod
    def from_enum_member(cls, current: enum.Enum, target: enum.Enum) -> typing.Self:
        """从枚举成员创建"""
        return cls(current.__class__.__name__, current.value, target.value)

    def dump_details_to_str(self) -> str:
        return f"cannot transist {self.__status_name} from {self.__current_status} to {self.__target_status}"


class Unauthorized(InternalError):
    """未认证"""

    def __init__(
        self, msg: Opt[str] = None, identity: Opt[str] = None, *args, **kwargs
    ):
        """
        :param identity: 导致未认证时使用的凭证（如果有）
        """
        super().__init__("you are using identity (token): \n%s" % identity)

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.UNAUTHORIZED


class Forbidden(InternalError):
    """无权操作"""

    def __init__(
        self,
        msg: str,
    ):
        super().__init__(msg)

    @property
    def task_status(self):
        return TaskStatus.FORBIDDEN


class TooManyRequests(InternalError):
    pass


# class UnavailableForLegalReasons(BlueFirmamentException):

#     def __init__(self, errmsg, *args, **kwargs) -> None:

#         super().__init__(451, "Unavailable For Legal Reasons, %s" % errmsg, *args, **kwargs)


# class TooManyRequests(BlueFirmamentException):

#     def __init__(self, errmsg, *args, **kwargs) -> None:

#         super().__init__(429, "Too many request, limit reached. %s" % errmsg, *args, **kwargs)



"""HTTP transport module, based on the ASGI specification.
"""

__all__ = [
    "HTTPTransporter"
]

import datetime
import typing
from typing import Optional as Opt
import urllib.parse
from dataclasses import dataclass
from .main import HTTPTransporter


@dataclass
class Cookie:
    name: str
    value: str
    path: Opt[str] = None
    domain: str = ''
    secure: bool = False
    '''如果为True，则只能在HTTPS协议下使用该Cookie'''
    httponly: bool = False
    '''JavaScript可否（False为可以）访问该Cookie'''
    expires: Opt[datetime.datetime] = None
    '''过期时间（datetime.datetime对象）'''
    max_age: int = 0
    '''最大存活时间（单位：秒）'''
    same_site: typing.Optional[typing.Union[typing.Literal['Lax'], typing.Literal['Strict']]] = None

    def dump_to_str(self) -> str:
        """Dump to string.
        
        Behaviours
        ----------
        - value will be url encoded
        """
        value: str = urllib.parse.quote(self.value, safe='')
        result = f'{self.name}={value};'
        if self.path:
            result += f' Path={self.path};'
        if self.domain:
            result += f' Domain={self.domain};'
        if self.secure:
            result += ' Secure;'
        if self.httponly:
            result += ' HttpOnly;'
        if self.expires:
            result += f' Expires={self.expires.strftime("%a, %d %b %Y %H:%M:%S GMT")};'
        if self.max_age:
            result += f' Max-Age={self.max_age};'
        if self.same_site:
            result += f' SameSite={self.same_site};'
        return result.strip()





"""Transporter listening to a Pub/Sub model channel.

TODO better logging
"""

__all__ = [
    "PubSubTransporter"
]

import typing

from ..exceptions import TaskHandlerNotFound
from ..transport.base import BaseTransporter
from ..task import Task, TaskResult
from ..dal.base import PubSubLikeDataAccessLayer, PubSubMessage

if typing.TYPE_CHECKING:
    from ..core import BlueFirmamentApp


class PubSubTransporter(BaseTransporter):

    def __init__(
        self,
        app: "BlueFirmamentApp",
        pubsub_dal: PubSubLikeDataAccessLayer,
        *channel_names: str,
        name: str = "default"
    ):
        """
        :param pubsub_dal: A PubSubDAL not subscribed to any channel.
        """
        super().__init__(app=app, name=name)

        self.__stop = False
        self.__pubsub_dal = pubsub_dal
        self.__channel_names = channel_names

    async def start(self):
        self.__stop = False
        await self.__pubsub_dal.subscribe(*self.__channel_names)
        self._logger.info("Listening to Pub/Sub channels", channels=self.__channel_names)
        async for message in self.__pubsub_dal.listen():
            try:
                await self(message)
            except TaskHandlerNotFound as e:
                self._logger.warning("No handler found for the task", task=e.task_id)
            except Exception as e:
                self._logger.exception(
                    f"Unknown error occured when handling task from Pub/Sub {self.name}"
                )
            if self.__stop:
                break

    async def stop(self):
        self.__stop = True
        await self.__pubsub_dal.unsubscribe(*self.__channel_names)
        self._logger.info("Stop listening to Pub/Sub channels", channels=self.__channel_names)

    async def __call__(self, message: PubSubMessage):
        await self._app.handle_task(
            task=Task.load_from_bytes(message["data"]),
            task_result=TaskResult(),
            transporter=self
        )




# author: Lan_zhijiang
# desc: quick imports

import importlib
import pkgutil

# By Google Gemini
def import_modules(package_name, var_name):
    """
    导入指定包下的所有模块并提取`<var_name>`变量

    如果找不到某模块中的变量，会被忽略

    @params
        package_name: 包名，例如 "my_project.routers"

    @return
        所有*_router*变量的列表
    """
    modules = []
    for module in pkgutil.iter_modules([package_name]):
        # 过滤非包模块
        if not module.ispkg:
            module_name = f"{package_name.replace('/', '.')}.{module.name}"
            module = importlib.import_module(module_name)
            # 提取*endswith*变量
            for attr in dir(module):
                try:
                    target_var = getattr(module, attr)
                except AttributeError:
                    continue
                else:
                    modules.append(target_var)
    return modules




"""Some basic types for the blue_firmament package.
"""

__all__ = [
    'AnnotatedDirective',
    'Undefined', '_undefined',
    "NamedTupleTV",
    "PathParamsT",
    "TaskRegistriesT",
    "CallableTV",
]

import enum
import typing

if typing.TYPE_CHECKING:
    from .transport.base import BaseTransporter
    from .task import TaskRegistry


class AnnotatedDirective(enum.Enum):
    """Directive used in typing.Annotated arguments.
    """
    NOLOG = 1
    """Don't log this value.
    """


class Undefined(enum.Enum):
    token = 'undefined'
_undefined: typing.Final = Undefined.token

NamedTupleTV = typing.TypeVar("NamedTupleTV", bound=typing.NamedTuple)

type PathParamsT = typing.Dict[str, typing.Any]
"""Path parameters type.

Path parameters is the parameters resolved from TaskID path.
"""

type TaskRegistriesT = dict["BaseTransporter" | str, "TaskRegistry"]
"""A dict, records which task registry serves which transporter.
"""

CallableTV = typing.TypeVar("CallableTV", bound=typing.Callable)



import abc
import copy
import enum
import typing
from typing import Optional as Opt

from . import DALQueryComponent
from ..utils import dump_field_like
from ..types import FieldLikeType, DALPath
if typing.TYPE_CHECKING:
    from ...scheme.field import Field


class DALFilter(DALQueryComponent):

    __use_fqfn__ = False

    def __init__(self, field: FieldLikeType, value: typing.Any):
        self._field = field
        self._value = value

    @property
    def dal_path(self) -> Opt[DALPath]:
        from ...scheme.field import Field, FieldValueProxy
        if isinstance(self._field, Field):
            return self._field.scheme_cls.dal_path()
        if isinstance(self._field, FieldValueProxy):
            return self._field.scheme.dal_path()
        return None

    @property
    def field_name(self) -> str:
        """Get field name.

        If `__use_fqfn__` is True, return fully qualified field name.
        """
        field_name = dump_field_like(self._field)
        if self.__use_fqfn__ and self.dal_path:
            return f"{self.dal_path[0]}.{field_name}"
        return field_name

    def _dump_value(self, val: typing.Any):
        from ...scheme.field import Field, FieldValueProxy
        if isinstance(self._field, Field):
            return self._field.dump_val_to_jsonable(val)
        if isinstance(self._field, FieldValueProxy):
            return self._field.field.dump_val_to_jsonable(val)
        if isinstance(val, enum.Enum):
            return val.value
        return val

    @property
    def value(self):
        """Get primitive python types of value

        - enum.Enum -> enum.Enum.value
        """
        return self._dump_value(self._value)

    def fork(
        self,
        use_fqfn: bool = False
    ) -> typing.Self:
        forked = copy.copy(self)
        forked.__use_fqfn__ = use_fqfn
        return forked

    def __repr__(self):
        return super().__repr__() + f'(field={self._field},value={self.value})'


class EqFilter(DALFilter):
    
    def dump_to_sql(self) -> str:
        return f"{self.field_name} = {repr(self.value)}"

    def dump_to_postgrest(self):
        value = self.value
        if isinstance(value, (list, tuple)) and len(value) == 0:
            return ('eq', (self.field_name, '{}'))
        return ('eq', (self.field_name, value))

    def dump_to_postgrest_str(self) -> str:
        value = self.value
        if isinstance(value, typing.Iterable):
            value = tuple(value)
        return f"{self.field_name}.eq.{value}"

class NotEqFilter(DALFilter):
    
    def dump_to_sql(self) -> str:
        return f"{self.field_name} != {repr(self.value)}"

    def dump_to_postgrest(self):
        return ('neq', (self.field_name, self.value))

    def dump_to_postgrest_str(self) -> str:
        return f"{self.field_name}.neq.{repr(self.value)}"

class IsFilter(DALFilter):

    __filter_name__ = 'is_'
    
    def __init__(self, field: FieldLikeType, value: bool | None) -> None:
        super().__init__(field, value)
    
    def dump_to_sql(self) -> str:
        return f"{self.field_name} IS {repr(self.value)}"

    def dump_to_postgrest(self):
        return (self.__filter_name__, (self.field_name, self.value))

    def dump_to_postgrest_str(self) -> str:
        if self.value is None:
            return f"{self.field_name}.is.null"
        elif self.value is True:
            return f"{self.field_name}.is.true"
        elif self.value is False:
            return f"{self.field_name}.is.false"
        else:
            raise ValueError("Value must be bool or None for IsFilter")

class InFilter(DALFilter):

    def __init__(self, field: FieldLikeType, value: typing.Iterable) -> None:
        super().__init__(field, value)

    @property
    def value(self):
        return tuple(
            self._dump_value(i)
            for i in self._value
        )
        # Dump every element. \
        # Since column must be element type to use InFilter.
    
    def dump_to_sql(self) -> str:
        return f"{self.field_name} IN ({', '.join(str(v) for v in self.value)})"

    def dump_to_postgrest(self):
        return ("in_", (self.field_name, self.value))

    def dump_to_postgrest_str(self) -> str:
        return f"{self.field_name}.in.({','.join(str(v) for v in self.value)})"

class ContainsFilter(DALFilter):
    
    def __init__(self, field: FieldLikeType, /, *values: typing.Any) -> None:
        super().__init__(field, values)
    
    def dump_to_sql(self) -> str:
        return f"{self.field_name} @> ARRAY[{', '.join(repr(v) for v in self.value)}]"

    def dump_to_postgrest(self):
        return ("contains", (self.field_name, self.value))

    def dump_to_postgrest_str(self) -> str:
        values = ",".join(f'"{value}"' for value in self.value)
        return f"{self.field_name}.cs.{{{values}}}"



"""BlueFirmament Utils Module.

Contents:
- `typing_`: Utilities related to type checking and type hinting.
- `dict_`: Utilities for enhanced dictionary operations.
- `enum_`: Utilities for handling enumerations.
- `json_`: Utilities for JSON serialization and deserialization.
- `enum_`
- `inspect_`
- `exec_`
- `datetime_`
- `encryption`
"""


"""Task
"""

__all__ = [
    "TaskID",
    "TaskMetadata",
    "Task",
    "LazyParameter",
    "Method"
]

import abc
import dataclasses
import enum
import json
import types
import uuid
import typing
from typing import Optional as Opt

from ..utils.enum_ import dump_enum, load_enum
from .._types import PathParamsT, Undefined, _undefined
from ..scheme.converter import AnyConverter, get_converter_from_anno

if typing.TYPE_CHECKING:
    from ..scheme.converter import BaseConverter


class Method(enum.Enum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    PATCH = 'PATCH'
    DELETE = 'DELETE'
    OPTIONS = 'OPTIONS'


class TaskID:
    """Identifier of Task.

    Consist of a method and a path.
    Path is segmented by seperator, which optimizing
    performance, enhancing naming clarity.
    Path can have variable part, which is path parameter.
    Path parameter is named, match by position.
    Path parameter have type annotation, helps
    ensures parameters type safety.

    :ivar __param_converters: Map of parameter name to converter
        Contains all parameters' converters.
    """

    def __init__(
        self,
        method: Opt[Method | str],
        path: str,
        separator: str = '/',
        param_types: Opt[typing.Mapping[str, typing.Type]] = None,
        param_converters: Opt[dict[str, "BaseConverter"]] = None
    ):
        """
        :param method:
            Action performed on the resource that path indicates.
            None is wildcard
        :param path: A URL-like path.
            Can include named positional parameters
            (e.g., '/users/{id}', 'users', 'users/').
        :param separator: Separator to split path into segments. Defaults to '/'.
        :param param_types: Path parameter's types.
            If not provided, all parameters will use AnyConverter.
            If conversion fails (value type safety failed), it's not considered a match.
        :param param_converters: Converter for path parameters.
            If not provided, inferred from param_types.
        """
        self.__method: Method = load_enum(Method, method)
        self.__path: str = '/' + path.replace(separator, '/').strip('/')
        """Normalized path.
        
        - must start with a slash
        """

        self.__segments: typing.List[str] = path.strip(separator).split(separator)
        '''Path segmented by separator'''
        self.__dynamic_indices: typing.List[int] = []
        '''Dynamic segments(path parameters) indices'''
        for i, segment in enumerate(self.__segments):
            if segment.startswith('{') and segment.endswith('}'):
                # Dynamic segment, remove the brackets
                self.__segments[i] = segment[1:-1]
                self.__dynamic_indices.append(i)

        self.__param_types: typing.Mapping[str, typing.Type] = param_types or {}
        self.__param_converters: typing.Dict[str, BaseConverter]
        if not param_converters:
            self.__param_converters = {}
            for dynamic_index in self.__dynamic_indices:
                param_name = self.__segments[dynamic_index]
                param_type = self.__param_types.get(param_name, _undefined)
                if param_type is _undefined:
                    self.__param_converters[param_name] = AnyConverter()
                else:
                    self.__param_converters[param_name] = get_converter_from_anno(param_type)
        else:
            self.__param_converters = param_converters

    @staticmethod
    def resolve_dynamic_indices(
        raw_path: str,
        remove_param_indicator: bool = True
    ) -> tuple[str, ...]:
        """
        :param remove_param_indicator:
            Remove curly brackets which used to mark a parameter in
            raw path.
        """
        return tuple(
            i[1:-1] if remove_param_indicator else i
            for i in raw_path.strip('/').split('/')
            if i.startswith('{') and i.endswith('}')
        )

    def __eq__(self, other):
        """Is two TaskID match.
        """
        if not isinstance(other, TaskID):
            return False
        return self.is_match(other) is not None

    def __hash__(self):
        return hash((self.method, *(i for i in self.segments)))

    def __str__(self):
        return f"{dump_enum(self.method) or ""}@{self.path}"

    def __len__(self) -> int:
        return len(self.segments)

    def __getitem__(self, key) -> 'TaskID':
        """使用slice获取子路由键
        """
        if isinstance(key, slice):
            return TaskID(self.__method, '/'.join(self.__segments[key]), param_converters=self.__param_converters)
        else:
            return TaskID(self.__method, self.__segments[key], param_converters=self.__param_converters)

    def dump_to_str(self) -> str:
        return self.__str__()

    @classmethod
    def load_from_str(cls, raw: str) -> typing.Self:
        method, path = raw.split("@", maxsplit=1)
        return cls(
            method=load_enum(Method, method or None),
            path=path,
        )

    def __is_segment_match(
        self,
        to_match_segment: str,
        my_segment_index: int
    ) -> types.NoneType | Undefined | typing.Any:
        """Whether a segment match another segment
        and resolve parameter if dynamic.

        :returns:
            If not match, return _undefined.
            If match and dynamic segment, return parameter value (converted).
            If match and static segment, return None.
        """
        if my_segment_index in self.__dynamic_indices:  # dynamic segment
            param_name = self.segments[my_segment_index]
            try:
                param_converter = self.__param_converters[param_name]
                return param_converter(to_match_segment)
            except KeyError:
                raise ValueError(f"No converter for parameter: {param_name}")
            except ValueError:
                return _undefined
        else:  # static segment
            return None if self.segments[my_segment_index] == to_match_segment else _undefined

    def fork(
        self,
        path_prefix: str = ""
    ) -> typing.Self:
        """Fork a TaskID

        :param path_prefix:
            Added to the front of the original path.
        :return: a new TaskID instance.
        """
        return TaskID(
            method=self.__method,
            path=f"{path_prefix}{self.__path}",
            param_types=self.__param_types
        )

    @property
    def method(self) -> typing.Optional[Method]:
        return self.__method

    @property
    def path(self) -> str:
        """Normalized path.
        """
        return self.__path

    @property
    def segments(self) -> typing.List[str]:
        return self.__segments

    def is_dynamic(self, allow_method_dynamic: bool = False) -> bool:
        """Is the TaskID dynamic.

        :return: If True, it's dynamic, otherwise static.
            Dynamic means it has path parameters or method is None (wildcard).
        """
        return len(self.__dynamic_indices) != 0 or (
            self.__method is None and not allow_method_dynamic
        )

    def resolve_params(self, segments: list[str]) -> dict[str, typing.Any]:
        """Resolve path parameters from (static) segments.

        Returns
        ^^^
        - 字典键一定包含本路由键定义的所有参数
        - 如果参数校验不通过或不存在，则为 _undefined
        """
        result = {}

        for i in self.__dynamic_indices:
            param_name = self.__segments[i]
            try:
                param_converter = self.__param_converters[param_name]
                result[param_name] = param_converter(segments[i])
            except (IndexError, KeyError, ValueError):
                # IndexError: 我方不存在对应的分段
                # KeyError: 没有对应的参数校验器
                # ValueError: 校验不通过
                result[param_name] = _undefined

        return result

    def is_match(
        self,
        to_match: 'TaskID',
    ) -> Opt[PathParamsT]:
        """Whether another task_id match this task_id.

        :param to_match: TaskID to match.

        :returns: Path parameters when matched, otherwise None.

        Behaviours
        ^^^^^^^^^^
        - Is method match. If method is None(wildcard), skip.
        - Is path match.
            - If strict, length must match.
            - If path is static，比较我方路径分段（仅静态）是否为对方的子集（严格模式比较路径字符串）
            - 如果有路径参数，比较我方路径分段（包括静态与动态）是否为对方的子集（严格模式检验对方是否为我方子集）
                - 此处解析路径参数，如果未找到或校验不通过，则不记录
        """
        if not isinstance(to_match, TaskID):
            return None

        if self.method:
            if self.method != to_match.method:
                return None

        if len(self.segments) != len(to_match.segments):
            return None

        if len(self.__dynamic_indices) == 0:
            # no dynamic segments, so we can compare segments directly
            return {} if self.segments == to_match.segments else None

        params = {}
        for i in range(len(to_match.segments)):
            seg_match = self.__is_segment_match(to_match.segments[i], i)
            if seg_match is _undefined:
                return None
            else:
                if seg_match is not None:
                    params[self.segments[i]] = seg_match

        return params


@dataclasses.dataclass
class TaskMetadata:
    """BlueFirmament Task Metadata
    """

    authorization: Opt[tuple[str, str]] = None
    """tuple(type, credentials)"""
    trace_id: Opt[str] = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    client_id: Opt[str] = None
    state: dict = dataclasses.field(default_factory=dict)
    """cookies"""

    def dump_to_bytes(
        self,
        encoding: str = 'latin-1',
        extra: Opt[dict] = None
    ) -> typing.Iterable[tuple[bytes, bytes]]:
        """Dump to iterable of (key, value) tuples, which key and value are bytes.

        :param extra: Extra data(in dict) to dump.
        """
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if value is not None:
                yield field.name.encode(encoding), str(value).encode(encoding)
        if extra:
            for key, value in extra.items():
                yield key.encode(encoding), str(value).encode(encoding)

    def dump_to_dict(self) -> dict[str, typing.Any]:
        return {
            field.name: getattr(self, field.name)
            for field in dataclasses.fields(self)
        }


class LazyParameter(abc.ABC):

    @abc.abstractmethod
    async def get(self) -> typing.Any:
        ...


TV = typing.TypeVar("TV")
class TaskParameters:

    def __init__(self, **parameters: typing.Any | LazyParameter):
        self.__parameters = parameters

    def __getitem__(self, item: str | enum.Enum):
        value = self.__parameters[dump_enum(item)]
        if isinstance(value, LazyParameter):
            raise TypeError("Use get() for LazyParameter")
        return value

    def items(self):
        return self.__parameters.items()

    async def get(self, item: str, default: TV = None) -> typing.Any | TV:
        value = self.__parameters.get(item, default)
        if isinstance(value, LazyParameter):
            return await value.get()
        return value

class Task:
    """Transport Task

    Task will be handled by handler and returns a result.

    Notes
    -----
    - Path parameters is not part of Task, cause task is created from transport layer,
    and path parameters are resolved by application layer. (This implies that task module
    is the glue layer between transport and application layers.)
    """

    def __init__(
        self,
        task_id: TaskID,
        metadata: Opt[TaskMetadata | dict] = None,
        parameters: Opt[dict[str, LazyParameter | typing.Any] | TaskParameters] = None,
    ) -> None:
        self.__task_id = task_id

        self.__parameters: TaskParameters
        if isinstance(parameters, dict):
            self.__parameters = TaskParameters(**parameters)
        elif isinstance(parameters, TaskParameters):
            self.__parameters = parameters
        else:
            self.__parameters = TaskParameters()

        if isinstance(metadata, dict):
            self.__metadata = TaskMetadata(**metadata)
        else:
            self.__metadata: TaskMetadata = metadata or TaskMetadata()

    @property
    def id(self) -> TaskID:
        return self.__task_id
    @property
    def trace_id(self) -> str:
        return self.__metadata.trace_id
    @property
    def metadata(self) -> TaskMetadata:
        return self.__metadata
    @property
    def parameters(self) -> TaskParameters:
        return self.__parameters

    async def dump_to_bytes(self, encoding: str = "utf-8") -> bytes:
        return json.dumps({
            "task_id": self.__task_id.dump_to_str(),
            "metadata": self.__metadata.dump_to_dict(),
            "parameters": {
                key: value if not isinstance(value, LazyParameter) else await value.get()
                for key, value in self.__parameters.items()
            }
        }).encode(encoding)

    @classmethod
    def load_from_bytes(cls, raw: bytes, encoding: str = "utf-8") -> typing.Self:
        data = json.loads(raw.decode(encoding))
        task_id = TaskID.load_from_str(data["task_id"])
        metadata = TaskMetadata(**data["metadata"])
        parameters = TaskParameters(**data["parameters"])
        return cls(task_id, metadata, parameters)



__all__ = [
    "BaseManager",
    # "BaseFieldManager"
]

import abc
import typing
from typing import Optional as Opt

from .._types import TaskRegistriesT
from ..transport.base import BaseTransporter
from ..task.registry import TaskRegistry, TaskEntry
from ..task.context import BaseTaskContext
from ..exceptions import BFExceptionTV
from ..scheme.field import Field
from ..scheme import SchemeTV
from ..log import log_manager_handler


class ManagerMetaclass(abc.ABCMeta):
    """Metaclass of Manager

    Terms
    -----
    Handlers are the methods excluding protected, private,
    classmethod, and staticmethod.

    Features
    --------
    Log enhancement
    ^^^^^^^^^^^^^^^
    Use :func:`blue_firmament.log.decorators.log_manager_handler` to decorate all handlers.
    A decorated handler will be skipped.

    Task registry
    ^^^^^^^^^^^^^
    Handlers decorated with :meth:`blue_firmament.task.task` will be
    automatically added to manager task registry.

    """

    def __new__(
        cls,
        name: str,
        bases: tuple[type[typing.Any], ...],
        attrs: dict[str, typing.Any],
        path_prefix: str = "",
        **kwargs,
    ):
        # exclude BaseManager
        if name in ("BaseManager",):
            return super().__new__(cls, name, bases, attrs, **kwargs)

        attrs["__path_prefix__"] = path_prefix
        attrs["__task_registries__"]: dict[BaseTransporter | str, TaskRegistry] = {}
        task_entries: list[tuple[tuple[BaseTransporter | str], TaskEntry]] = []

        for attr_name, attr_value in attrs.items():
            # resolve task_entries
            if (
                isinstance(attr_value, tuple)
                and len(attr_value) == 2
                and isinstance(attr_value[0], tuple)
                and isinstance(attr_value[1], TaskEntry)
            ):
                task_entry = attr_value[1]
                entry_handlers = task_entry.handlers
                # unwrap handler
                if len(entry_handlers) != 1:
                    raise ValueError("Must have exactly one handler")
                attrs[attr_name] = entry_handlers[0].function
                # add to entries
                task_entries.append(attr_value)
                # make later resolution works
                attr_value = attrs[attr_name]

            if attr_name.startswith("_"):
                continue

            # log enhancement
            if callable(attr_value):
                if not isinstance(attr_value, (classmethod, staticmethod)):
                    attrs[attr_name] = log_manager_handler(attr_value)

        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)

        if not issubclass(new_cls, BaseManager):
            raise TypeError(f"{name} should not directly use the ManagerMetaclass")

        # set task handlers' manager class
        for i in task_entries:
            i[1].set_manager_cls(new_cls)
            for transporter in i[0]:
                attrs["__task_registries__"].setdefault(
                    transporter,
                    TaskRegistry(name=str(transporter), path_prefix=path_prefix),
                ).add_entry(i[1])

        return new_cls


T = typing.TypeVar("T")


class BaseManager(
    typing.Generic[SchemeTV],
    BaseTaskContext,
    metaclass=ManagerMetaclass,
):
    """Base class of manager.

    Configuration
    -------------
    Config through bases parameters.
    """

    __scheme_cls__: type[SchemeTV]
    """Scheme class this manager is managing
    """
    __task_registries__: TaskRegistriesT
    """Task entries of this manager
    """
    __manager_name__: str
    """Friendly name of this manager.

    - no ``manager``
    - use ``_`` and lowercase
    """
    __path_prefix__: str

    class ManagingScheme:
        def __init__(self, scheme: Opt[SchemeTV] = None):
            self.__scheme: Opt[SchemeTV] = scheme

        def set(self, scheme: SchemeTV):
            self.__scheme = scheme

        def get(self) -> SchemeTV:
            return self.__scheme

        def __bool__(self):
            return self.__scheme is not None

    def __init_subclass__(
        cls,
        scheme_cls: Opt[type[SchemeTV]] = None,
        manager_name: Opt[str] = None,
        **kwargs,
    ):
        if scheme_cls:
            cls.__scheme_cls__ = scheme_cls
        if manager_name:
            cls.__manager_name__ = manager_name

        super().__init_subclass__(**kwargs)

    def __init__(self, task_context: BaseTaskContext) -> None:
        BaseTaskContext.__init__(self, task_context)

        self.__scheme = self.ManagingScheme(None)
        self._logger = self._logger.bind(manager_name=self.__manager_name__)
        self.__post_init__()

    def __post_init__(self):
        """Override this method to customize init behaviour

        This method will be called once BaseManager finish its init.
        """
        ...

    @property
    def _scheme_cls(self) -> typing.Type[SchemeTV]:
        """Managing scheme class"""
        return self.__scheme_cls__

    @property
    def _dal_path(self):
        """DALPath of managing scheme"""
        return self._scheme_cls.dal_path()

    @property
    def _scheme_key(self) -> Field:
        """Key field of managing scheme"""
        return self._scheme_cls._get_key_field()

    @property
    def _scheme(self) -> SchemeTV:
        """Managing scheme

        :raise ValueError: scheme not set
        """
        if not self.__scheme:
            raise ValueError("scheme is not set")
        return self.__scheme.get()

    @_scheme.setter
    def _scheme(self, scheme: SchemeTV):
        """Set managing scheme"""
        self.__scheme.set(scheme)

    def _try_get_scheme(self) -> Opt[SchemeTV]:
        """Get scheme without exception"""
        try:
            return self._scheme
        except ValueError:
            self._logger.warning("Scheme not set")
            return None

    def _get_bfe(
        self, exception: typing.Type[BFExceptionTV], *args, **kwargs
    ) -> BFExceptionTV:
        """获取碧霄异常实例

        携带和日志器一致的上下文信息
        """

        return exception(*args, **kwargs, **self._logger._context)


# class ManagerRouteRegister:

#     """管理器路由注册器

#     用于将管理器中的处理器注册到 App 的根路由器中。
#     """

#     def __init__(self,
#         app: "BlueFirmamentApp",
#         manager_cls: typing.Type[BaseManager],
#         use_manager_prefix: bool = True
#     ) -> None:

#         self._app = app
#         self._manager_cls = manager_cls
#         self._use_manager_prefix = use_manager_prefix
#         self._register = self._app.router.get_manager_route_resigter(
#             manager=self._manager_cls,
#             use_manager_prefix=self._use_manager_prefix
#         )

#     def __enter__(self) -> typing.Self:
#         return self

#     def __exit__(self,
#         exc_type: typing.Optional[typing.Type[BaseException]],
#         exc_val: typing.Optional[BaseException],
#         exc_tb: typing.Optional[types.TracebackType]
#     ) -> None:
#         # No specific cleanup needed for now
#         pass

#     def __call__(self,
#         operation: Method,
#         path: str,
#         handler: typing.Callable,
#     ):

#         """注册一个路由记录
#         """
#         return self._register(operation, path, handler)


# BaseManagerTV = typing.TypeVar("BaseManagerTV", bound=BaseManager)
# class BaseFieldManager(
#     typing.Generic[
#         T,
#         BaseManagerTV,
#         SessionTV
#     ],
#     BaseTaskContext[SessionTV],
#     metaclass=ManagerMetaclass,
# ):
#     """字段管理器基类

#     Example
#     -------
#     .. code-block:: python
#         from blue_firmament.scheme import BlueFirmamentField, BaseScheme, BaseManager
#         from blue_firmament.manager import BaseFieldManager

#         class MessageContent(BlueFirmamentField):
#             def __init__(self):
#                 super().__init__(default="Hello World")

#         class Message(BaseScheme):
#             content: MessageContent

#         class MessageManager(BaseManager[Message, CommonSession]):
#             __scheme_cls__ = Message

#         class MessageContentManager(BaseFieldManager[MessageContent, CommonSession]):
#             __field_cls__ = MessageContent
#             __scheme_manager_cls__ = MessageManager
#     """

#     __scheme_manager_cls__: typing.Type[BaseManagerTV]
#     """所管理字段所属数据模型的管理器类
#     """
#     __field__: Field[T]
#     """管理器管理的字段类
#     """
#     __path_prefix__: str
#     """管理器路径前缀
#     """
#     __manager_name__: str = "UnknownFieldManager"
#     """管理器友好名称
#     """

#     def __init__(self, request_context: BaseTaskContext) -> None:

#         self.init_from_tc(request_context)

#         self._field_value: Opt[T] = None
#         """管理的字段值"""
#         self.__scheme_manager: BaseManagerTV = self.__scheme_manager_cls__(request_context)
#         self.__logger: LoggerT = self._logger.bind(
#             name=self.__manager_name__
#         )
#         """管理器级别日志记录器"""

#     @property
#     def _logger(self) -> LoggerT:

#         """管理器级别日志记录器
#         """
#         return self.__logger

#     @property
#     def field(self) -> Field[T]:
#         """管理的字段实例"""
#         return self.__field__

#     @property
#     def scheme_manager(self) -> BaseManagerTV:
#         """管理的字段所属数据模型的管理器实例"""
#         return self.__scheme_manager

#     @property
#     def value(self) -> T:
#         """当前管理的字段值"""
#         if not self._field_value:
#             raise ValueError('field value is None')
#         return self._field_value

#     @classmethod
#     def get_route_register(cls, app: "BlueFirmamentApp") -> "ManagerRouteRegister":
#         """获取路由注册器
#         """
#         return ManagerRouteRegister(
#             app=app,
#             manager_cls=cls,  # TODO type safe
#             use_manager_prefix=True
#         )



"""Blue Firmament's abilities that helps you defining schemas for your backend application."""


__all__ = [
    "BaseScheme", "NoProxyScheme", "SchemeTV", "merge_scheme",
    "BusinessScheme", "EditableScheme",
    "field", "private_field", "FieldT", "PFieldT", "CompositeField",
    "field_validator", "field_validators", "scheme_validator",
    "BaseConverter", "AnyConverter", "UnionConverter", "OptionalConveter",
    "StrConverter", "IntConverter",
    "ListConverter", "TupleConverter", "SetConverter", "DictConverter", 
    "DatetimeConverter", "TimeConverter",
]

from .main import BaseScheme, NoProxyScheme, SchemeTV, merge as merge_scheme
from .business import (
    BusinessScheme, EditableScheme
)
from .field import (
    field, private_field, 
    Field as FieldT,
    PrivateField as PFieldT,
    CompositeField
)
from .converter import (
    BaseConverter, 
    AnyConverter, UnionConverter, OptionalConveter,
    StrConverter, IntConverter,
    ListConverter, TupleConverter, SetConverter,
    DictConverter,
    DatetimeConverter, TimeConverter,
)
from .validator import (
    field_validator, field_validators, scheme_validator
)



"""BlueFirmament event system.
"""

__all__ = [
    "set_event_broker",
    "Event",
    "emit",
    "simple_emit"
]

import typing
from typing import Annotated as Anno, Optional as Opt, Literal as Lit

from .log import get_logger
from .task import TaskID, Task
from .dal.base import PubSubLikeDataAccessLayer

LOGGER = get_logger(__name__)


EVENT_BROKER: PubSubLikeDataAccessLayer
"""A PubSubDAL.

- must configured a default channel. 
"""

def set_event_broker(event_broker: PubSubLikeDataAccessLayer):
    global EVENT_BROKER
    if not isinstance(event_broker, PubSubLikeDataAccessLayer):
        raise TypeError("event_broker must be a PubSubLikeDataAccessLayer instance")
    EVENT_BROKER = event_broker


class Event(Task):
    """Event is a specialized Task.
    """


async def emit(event: Event) -> None:
    """Emit an event to the event broker.
    """
    await EVENT_BROKER.publish(await event.dump_to_bytes())
    LOGGER.info("Event emitted", event_id=event.id)

def simple_emit(
    name: str,
    parameters: Opt[dict] = None,
    metadata: Opt[dict] = None,
) -> typing.Coroutine[None, None, None]:
    """Emit an event in a simple way.

    :param name: Name of the event.
        e.g. "user.created", "order.completed"
    :param parameters: Parameters of the event.
    :param metadata: Metadata of the event.
        Fields must be defined in :meth:`blue_firmament.task.TaskMetadata`
    """
    event = Event(
        task_id=TaskID(method=None, path=name, separator='.'),
        parameters=parameters,
        metadata=metadata
    )
    return emit(event)




"""
author: Lan_zhijiang
date: 2024-11-10
desc: 时间日期工具函数
issues: 

references: 
    https://git.hadream.ltd/anana/backend/common/-/wikis/Util/Datetime
"""

import datetime
import time

# settings
from ..data.settings.base import get_setting as get_base_setting


def get_timezone(timezone_delta: int = get_base_setting().timezone_delta) -> datetime.timezone:

    """
    获取时区对象
    """
    return datetime.timezone(datetime.timedelta(hours=timezone_delta))

def get_datetimez(
    rfc3339: str | None = None,
    timestamp: float | None = None,
    iso8601: str | None = None,
    timezone: datetime.timezone = get_timezone(),
) -> datetime.datetime:

    """
    获取当前时间对象（带时区）

    :param rfc3339: RFC3339格式时间字符串，如"2024-11-10T12:00:00+08:00"
    :param timestamp: 时间戳（秒级浮点）；如果使用此，时区固定为 UTC+0
    :param iso8601: ISO8601格式时间字符串，如"2024-11-10 12:00:00"
    :param timezone: 时区对象，可以通过 `get_timezone(8)` 获取东八区时区
    :raise ValueError: 如果rfc3339或iso8601都未提供，则抛出异常
    """
    if timestamp:
        return datetime.datetime.fromtimestamp(timestamp, tz=get_timezone(0))
    if rfc3339 or iso8601:
        val = rfc3339 or iso8601
        if val is None:
            raise ValueError("rfc3339 or iso8601 must be provided")
        return datetime.datetime.fromisoformat(val)
    return datetime.datetime.now(tz=timezone)

def get_datetime(timestamp: float | None = None) -> datetime.datetime:

    """
    获取当前时间对象（不带时区）
    """
    if timestamp:
        return datetime.datetime.fromtimestamp(timestamp)
    return datetime.datetime.now()

def get_timestamp(datetime_obj: datetime.datetime | None = None, rfc3339: str | None = None) -> float:

    """
    获取时间戳（秒级浮点）
    """
    if datetime_obj:
        return datetime_obj.timestamp()
    if rfc3339:
        return datetime.datetime.fromisoformat(rfc3339).timestamp()
    return time.time()

def get_rfc3339(timestamp: float | None = None, datetime_obj: datetime.datetime | None = None) -> str:

    """
    获取RFC3339格式时间
    """
    if timestamp:
        return datetime.datetime.fromtimestamp(timestamp).isoformat()
    elif datetime_obj:
        return datetime_obj.isoformat()
    
    raise ValueError("timestamp or datetime_obj must be provided")

def format_datetime(datetime: datetime.datetime, format: str = "%Y-%m-%d %H:%M:%S") -> str:

    """
    格式化时间对象
    """
    return datetime.strftime(format)

def format_timestamp(timestamp: float, format: str = "%Y-%m-%d %H:%M:%S") -> str:

    """
    格式化时间戳
    """
    return datetime.datetime.fromtimestamp(timestamp).strftime(format)


"""Utils for Authorization."""

__all__ = [
    "decode_jwt",
    "decode_token"
]

import jwt
import typing
from typing import Annotated as Anno, Optional as Opt, Literal as Lit
from ..data.settings.auth import get_setting as get_auth_setting
from ..exceptions import ParamsInvalid


class JWTPayload(typing.TypedDict):
    sub: str
    """Subject"""
    iss: str
    """Issuer"""
    iat: int
    """Issued at (Unix timestamp in seconds)"""
    exp: int
    """Expiration time (Unix timestamp in seconds)"""
    roles: typing.Iterable[str]
    """Roles of the subject"""
    sid: str
    """Session ID"""

def decode_jwt(
    token: str,
    key: str,
    algorithms: typing.Iterable[str] = ("HS256",),
    audience: typing.Iterable[str] = ()
) -> JWTPayload:
    try:
        return JWTPayload(**jwt.decode(
            jwt=token,
            key=key,
            algorithms=algorithms,
            audience=audience
        ))
    except jwt.exceptions.InvalidSignatureError:
        raise ParamsInvalid(
            'JWT signature invalid',
            token=token, key=key,
        )
    except jwt.exceptions.DecodeError:
        raise ParamsInvalid(
            'JWT decode failed',
            token=token, algorithm=algorithms,
        )


def decode_token(token: str, token_type: Lit["jwt"] | str) -> dict | JWTPayload:
    if token_type == "jwt":
        return decode_jwt(
            token=token,
            key=get_auth_setting().jwt_secret_key,
            algorithms=get_auth_setting().jwt_algorithms,
            audience=get_auth_setting().jwt_allowed_audiences
        )
    else:
        raise ParamsInvalid(f"Unsupported token type: {token_type}")




__all__ = [
    "get_logger", 
    "bind_logger_contextvars", "clear_logger_contextvars",
    "LoggerT",
    "log_manager_handler"
]

import structlog

from .main import (
    get_logger,
    bind_logger_contextvars, clear_logger_contextvars
)
from .decorators import (
    log_manager_handler
)

LoggerT = structlog.stdlib.BoundLogger




import json
import pkgutil


def load_json_file(
    file_path: str, encoding: str = "utf-8", package: str | None = None
) -> dict:

    """从文件系统加载JSON文件

    :param file_path: 文件路径
    :param encoding: 编码
    :param package: 包名；如果不为None，则使用pkgutil.get_data读取

    :return 转换为字典的JSON数据（异常返回空字典）
    """
    from ..log.main import get_logger
    logger = get_logger(__name__)
    
    try:
        if package:
            pkg_read_result = pkgutil.get_data(package, file_path)
            if pkg_read_result is None:
                logger.error('Load json file %s failed, file not found' % file_path)
                return {}
            data = json.loads(pkg_read_result.decode(encoding))
        else:
            data = json.load(open(file_path, "r", encoding=encoding))
        
        return data
    except IOError:
        logger.error('Load json file %s failed, IO error' % file_path)
        return {}
    except json.decoder.JSONDecodeError:
        logger.error('Load json file %s failed, unable to decode' % file_path)
        return {}

def save_json_file(file_path: str, data: dict, encoding: str = "utf-8"):

    json.dump(data, open(file_path, "w", encoding=encoding))



"""
国际化实用工具
"""

import typing
import enum

from ..log.main import get_logger
logger = get_logger(__name__)

# setting
from ..data.settings.i18n import get_setting

# utils
from .enum_ import dump_enum as get_enum_value

# libs
import gettext


LanguageEnum = Type[get_setting().language_enum]
DomainEnum = Type[get_setting().domain_enum]
DEFAULT_LANG = get_setting().default_lang
TRANSLATORS: Dict[LanguageEnum, Dict[DomainEnum, gettext.GNUTranslations]] = {}
'''翻译器实例，按照 lang, domain 存储'''


def get_translator(
    domain: DomainEnum | str, 
    language: LanguageEnum = DEFAULT_LANG, 
    localedir: str = None
) -> gettext.GNUTranslations:
    
    """
    获取翻译器实例

    :param domain: DomainEnum | str
        翻译领域
    :param language: LanguageEnum
        语言
    :param localedir: str
        自定义 locale 目录 \n
        如果没有传递，则使用配置中指定的目录 \n
        如果传递，且发现有重名的翻译器，则合并

    Usgae
    -----
    切换翻译器实例 ：
        通过 get_translator().install() 切换_()对应的翻译器实例或者在模块内自行声明

    Warning
    -------
    如果你希望合并翻译器，则必须先初始化默认翻译器实例，再获得自定义 localedir
    
    Docs
    ----
    https://git.hadream.ltd/anana/backend/gitlab-profile/-/wikis/Standard/Internationalization
    """
    customized_localedir = localedir is not None
    try:
        translator = TRANSLATORS[language][domain]
    except KeyError:
        if language not in TRANSLATORS:
            TRANSLATORS[language] = {}
    
        if domain not in TRANSLATORS[language]:
            if not localedir:
                localedir = get_setting().r_localedir
            
            translator = gettext.translation(
                get_enum_value(domain),
                localedir=localedir,
                languages=(get_enum_value(language), get_enum_value(DEFAULT_LANG))
            )
            TRANSLATORS[language][domain] = translator
    else:
        if customized_localedir:
            new_translator = gettext.translation(
                get_enum_value(domain),
                localedir=localedir,
                languages=(get_enum_value(language), get_enum_value(DEFAULT_LANG))
            )
            translator.add_fallback(new_translator)
    finally:
        return translator


def gettext_placeholder(message: str) -> str: 

    """
    gettext 占位符，用于标记需要翻译的字符串

    一般为`_`

    在翻译器实例安装后，会被替换为翻译后的字符串
    """
    return message


def pgettext_placeholder(context: str, message: str) -> str: 

    """
    pgettext 占位符，用于标记需要翻译的字符串，但带有场景值

    一般为`de_p`
    """
    return message


class TranslatableEnum(Enum):

    """
    可翻译 Enum

    在对 Enum 进行字符串格式化时，将成员值通过 i18n 翻译

    支持 context 模式，需要在指定`_context` \n
    设置成员时，使用 `_(<context>|<value>)`

    文档：https://git.hadream.ltd/anana/backend/common/-/wikis/Util/Enum#translatableenum
    """

    def __init__(self, 
        t: Optional[gettext.GNUTranslations] = None, 
        context: Optional[str] = None, 
        domain: Optional[DomainEnum] = None    
    ) -> None:
        
        """
        传入 t 或者 domain
        """

        self._translator = t
        self._context = context
        self._domain = domain

    @property
    def translator(self) -> gettext.GNUTranslations:
        if not self._translator:
            return get_translator(self._domain, DEFAULT_LANG)
        return self._translator

    def translate(self, lang: Optional[LanguageEnum] = None, domain: Optional[DomainEnum] = None) -> str:

        """
        如果配置了翻译器，则使用翻译器翻译成员值；否则直接返回成员值
        """
        if lang and domain:
            t = get_translator(domain, lang)
        else:
            t = self.translator

        if isinstance(t, gettext.GNUTranslations):
            if self._context:
                return t.pgettext(self._context, self._value_)
            return t.gettext(self._value_)
        else:
            return self._value_

    def __format__(self, *args) -> str:
        return self.translate()
    
    def __str__(self):
        return self.translate()



"""BlueFirmament Setting Module

"""

__all__ = [
    "Setting",
    "EnvSetting", 
    "JsonFileSetting",
    "EnvJsonSetting",
    "PythonScriptSetting",
    "make_setting_singleton",
    "private_field", "field",
]

import os
import pkg_resources
import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit
from .scheme.field import get_default
from .scheme import FieldT, field, private_field, BaseScheme
from .utils.file import load_json_file
from . import __name__ as PACKAGE_NAME


class Setting(BaseScheme,
    proxy=False,
):

    """Setting base class.
    """
    
    _setting_name: FieldT[str] = private_field()
    """配置名称"""
    _setting_path: FieldT[str | None] = private_field(default=None)
    """
    配置文件路径
    
    如果是打包在包内的配置文件，使用相对路径（不要用slash开头） \n
    如果是可以基于当前工作目录的配置文件，使用`./`或不使用任何前缀开头

    Example
    ^^^^^^^
    - ``./config.json``: 相对于CWD
    - ``data/config.json``: 相对于包
    - ``data/config/base``: 相对于包（目录）
    """
    _is_packaged: FieldT[bool] = private_field(default=True)
    """是否打包在包内"""
    _package_name: FieldT[str] = private_field(default=PACKAGE_NAME)
    """所属包的包名"""

    @property
    def is_packaged(self):
        return self._is_packaged
    
    @property
    def package_name(self):
        return self._package_name
    
    @classmethod
    def get_resource_path(cls, file_path: str) -> str:

        """
        获取被打包/未打包的资源的路径（免去判断资源是否被打包）

        :param file_path: str
            相对于该包的资源路径
        """
        if get_default(cls._is_packaged):
            return pkg_resources.resource_filename(
                get_default(cls._package_name),
                file_path
            )

        return file_path
    
    @property
    def setting_path(self):

        '''配置文件路径

        如果没有配置（None），直接返回 \n
        如果是打包在包内的配置文件，返回绝对路径 \n
        '''

        if self._setting_path is None:
            return None
        return self.get_resource_path(self._setting_path)

    @classmethod
    def load(cls) -> typing.Self:
        return cls()


class EnvSetting(
    Setting,
    partial=True
):

    """Load setting by environment.

    Kinds of environments:
    - base
    - local
    - your customized env, like production, development...

    Guidance
    --------
    Your base env setting class should inherit this class.
    And then define what field does this setting has in the base
    env setting class.

    Add more env like production, development or local (built-in) setting
    by inheriting your base env setting class. (This will ensure field's type
    safety)
    """

    __env_cls__: Opt[typing.List[typing.Type["EnvSetting"]]] = None

    def __init_subclass__(cls) -> None:
        if not cls.__env_cls__:
            cls.__env_cls__ = []
        cls.__env_cls__.append(cls)

    _env: FieldT[str] = private_field()

    @classmethod
    def __find_env_cls(cls, env_name: str) -> typing.Type["EnvSetting"]:
        if not cls.__env_cls__:
            raise ValueError("no env cls registered")
        for i in cls.__env_cls__:
            if get_default(i._env) == env_name:
                return i
        raise ValueError(f"env {env_name} setting not found")

    @classmethod
    def load(cls) -> typing.Self:
        base_cls = cls.__find_env_cls("base")
        local_cls = cls.__find_env_cls("local")
        env_cls = cls.__find_env_cls(os.environ.get("ENV", "production"))
        
        return typing.cast(typing.Self, 
            base_cls(**env_cls(**local_cls()))
        )


class JsonFileSetting(Setting):

    """
    JSON文件配置
    """

    @classmethod
    def load(cls) -> typing.Self:

        # 仍然需要使用 field_as_class_var，因为此时数据模型没有被实例化
        fp = get_default(cls._setting_path)
        if fp:
            data = load_json_file(cls.get_resource_path(fp))
            try:
                return cls(**data)
            except ValueError as e:
                from .log.main import get_logger
                logger = get_logger(__name__)
                logger.error(f"Validation error in JsonSettingLoader {get_default(cls._setting_path)}: {e}")
                raise e
        else:
            raise ValueError("Setting path is not set")

class EnvJsonSetting(Setting, partial=True):
    
    """
    多环境JSON配置
    """

    _setting_env: FieldT[str] = private_field(default_factory=lambda: os.environ.get("ENV", "production"))
    """配置环境"""

    @classmethod
    def load(cls) -> typing.Self:

        """
        加载多环境JSON配置文件

        顺序（优先级从低到高）：
        - .base.json
        - .{env}.json
        - .local.json

        可能抛出的错误：
        - ValidationError: 配置文件不符合数据模型定义
        """
        from .log.main import get_logger
        logger = get_logger(__name__)
        
        setting_name = get_default(cls._setting_name)
        setting_path = get_default(cls._setting_path)
        setting_env = get_default(cls._setting_env)
        package_name = get_default(cls._package_name) if get_default(cls._is_packaged) else None

        logger.debug(f"Loading setting {setting_name} in {setting_env} environment")

        data = {}
        data.update(load_json_file(f"{setting_path}/{setting_name}.base.json", package=package_name))
        data.update(load_json_file(f"{setting_path}/{setting_name}.{setting_env}.json", package=package_name))
        data.update(load_json_file(f"{setting_path}/{setting_name}.local.json", package=package_name))

        try:
            return cls(**data)
        except ValueError as e:
            from .log.main import get_logger
            logger = get_logger(__name__)
            logger.error(f"Validation error in {get_default(cls._setting_name)} EnvJsonSetting loader: {e}")
            raise e

class PythonScriptSetting(Setting):

    """
    Python脚本配置

    加载py脚本中的setting字段作为设置，该字段为可索引对象
    """

    def __init__(self):
        
        from .log.main import get_logger
        logger = get_logger(__name__)
        logger.debug(f"Loading python script setting: {self._setting_path}")

        py_setting = __import__(
            name=self._setting_name, fromlist=["setting"]
        ).setting
        super().__init__(**py_setting)


ClsType = typing.TypeVar("ClsType", bound=Setting)
def make_setting_singleton(
    cls_ins: ClsType
) -> typing.Tuple[typing.Callable[[], ClsType], typing.Callable[[ClsType], None]]:
    
    """使用单例模式来获取配置实例
    
    该函数返回两个函数：一个用于获取配置实例，一个用于设置配置实例

    Example
    ^^^^^^^

    .. code-block:: python
        from blue_firmament.setting import make_setting_singleton, Setting

        class MySetting(Setting):

            _setting_name = "my_setting"
            some_setting_field: str = 'a'
        
        get_setting, set_setting = make_setting_singleton(MySetting())
    """

    _setting_ins = cls_ins

    def get_setting():
        return _setting_ins
    
    def set_setting(setting: ClsType):
        nonlocal _setting_ins
        _setting_ins = setting

    return get_setting, set_setting


"""Utils of dict
"""


import enum
import typing
from typing import Annotated as Anno
from typing import Literal as Lit
from typing import Optional as Opt
from .enum_ import dump_enum


class EnhancedDict:
    def __init__(self, parameters):
        self.__parameters: typing.Dict[str, typing.Any] = parameters

    def get(self, key: str | enum.Enum, default: typing.Any = None):
        return self.__parameters.get(dump_enum(key), default)
    
    def set(self, key: str | enum.Enum, value: typing.Any):
        self.__parameters[dump_enum(key)] = value

    def __getitem__(self, key: str | enum.Enum):
        return self.__parameters[dump_enum(key)]



"""Common Session and its fields"""

import typing
import uuid
from typing import Annotated as Anno, Optional as Opt, Literal as Lit

from ..exceptions import Unauthorized
from ..utils import auth_
from ..auth import AuthSession, User
from ..dal.base import DataAccessObjects
from . import Session, SessionField
from ..log.main import get_logger

LOGGER = get_logger(__name__)


class AuthSessionField(SessionField[AuthSession | None]):
    class FromToken(typing.TypedDict):
        access_token: str
        access_token_type: str
        access_token_payload: dict
        refresh_token: Opt[str]

    def __init__(
        self,
        from_token: Opt[FromToken] = None,
    ):
        if from_token:
            super().__init__(
                AuthSession.from_token(
                    access_token=from_token["access_token"],
                    access_token_type=from_token["access_token_type"],
                    access_token_payload=from_token["access_token_payload"],
                    refresh_token=from_token.get("refresh_token", None),
                )
            )
        else:
            super().__init__(None)

    def is_expired(self) -> bool:
        if self.value:
            return self.value.is_expired()
        return False

    def refresh(self) -> None:
        if self.value:
            self.value.refresh()


class DAOsField(SessionField[DataAccessObjects]):
    def __init__(self, auth_session: AuthSession | AuthSessionField | None = None):
        if isinstance(auth_session, AuthSessionField):
            super().__init__(DataAccessObjects(auth_session.value))
        else:
            super().__init__(DataAccessObjects(auth_session))

    def is_expired(self) -> bool:
        return self.value.is_expired()


class CommonSession(Session):
    """BlueFirmament Common Session

    Configure by setting class variables.
    """

    ACCESS_TOKEN_TYPE: str = "jwt"
    ACCESS_TOKEN_PAYLOAD_ID_CLAIM: str = "sid"
    """which claim in access token payload will be used as session id.
    """
    __fields__ = ("auth_session", "daos")

    def __init_fields__(
        self,
        daos: SessionField[DataAccessObjects],
        auth_session: SessionField[AuthSession | None],
    ) -> None:
        self.__daos = daos
        self.__auth_session = auth_session

    @property
    def daos(self) -> DataAccessObjects:
        """DataAccessObjects"""
        return self.__daos.value

    @property
    def operator(self) -> User:
        return self.auth_session.user

    @property
    def auth_session(self) -> AuthSession:
        if self.__auth_session.value:
            return self.__auth_session.value.user
        raise Unauthorized("tries to access operator but unauthorized")

    @classmethod
    def from_task(cls, task) -> typing.Self:
        authorization = task.metadata.authorization
        if authorization:
            access_token = authorization[1]
        else:
            LOGGER.warning("authorization not found in metadata")
            access_token = None
        refresh_token = task.metadata.state.get("refresh_token", None)

        access_token_payload = {}
        if access_token:
            access_token_payload = auth_.decode_token(
                access_token, cls.ACCESS_TOKEN_TYPE
            )

        def get_fields():
            fields = {}
            if access_token and access_token_payload:
                fields["auth_session"] = AuthSessionField(
                    from_token=AuthSessionField.FromToken(
                        access_token=access_token,
                        access_token_type=cls.ACCESS_TOKEN_TYPE,
                        access_token_payload=access_token_payload,
                        refresh_token=refresh_token,
                    )
                )
            else:
                fields["auth_session"] = AuthSessionField()
            fields["daos"] = DAOsField(fields["auth_session"])
            return fields

        return cls.upsert(
            access_token_payload.get(
                cls.ACCESS_TOKEN_PAYLOAD_ID_CLAIM, uuid.uuid4().hex
            ),
            fields_getter=get_fields,
        )



"""Scheme for business
"""

__all__ = [
    "BusinessScheme",
    "EditableScheme"
]

import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit
from blue_firmament.dal.base import DataAccessLayer
from blue_firmament.dal.types import DALPath
from ..dal.types import KeyableType
from .field import field, Field as FieldT
from .main import BaseScheme, NoProxyScheme


KeyTV = typing.TypeVar('KeyTV', bound=KeyableType)
class BusinessScheme(
    typing.Generic[KeyTV],
    BaseScheme, 
):
    
    def __init_subclass__(
        cls,
        key_type: Opt[type[KeyTV]] = None,
        dump_flags: Opt[dict[str, set[str]]] = None,
        **kwargs
    ) -> None:
        super().__init_subclass__(**kwargs)
        if key_type:
            cls._id._set_converter_from_anno(key_type)
        if dump_flags:
            if "_id" in dump_flags:
                cls._id.dump_flags = dump_flags["_id"]

    _id: FieldT[KeyTV] = field(is_key=True, is_natural_key=True)

    @property
    def _inserted(self):
        """Is this instance inserted to P.DAL

        Behaviours
        ----------
        >>> BusinessScheme(_id=1)._inserted
        True
        >>> BusinessScheme(_id='abcd')._inserted
        True
        >>> BusinessScheme(_id=0)._inserted
        False
        >>> BusinessScheme(_id='')._inserted
        False
        >>> BusinessScheme(_id=CID(ida=1, idb=None))._inserted
        False
        >>> BusinessScheme(_id=CID(ida=1, idb='cc'))._inserted
        True
        """
        if isinstance(self._id, int):
            return self._id != 0
        if isinstance(self._id, str):
            return self._id != ''
        if isinstance(self._id, BaseScheme):
            return all(
                bool(i)
                for i in self._id.__field_values__.values()
            )
        return False


class EditableScheme(
    NoProxyScheme,
    partial=True,
    inherit_validators=True             
):
    pass




from ...setting import Setting, make_setting_singleton
from ...scheme import private_field, FieldT

class BaseSetting(Setting):

    _setting_name: FieldT[str] = private_field(default="base")

    timezone_delta: int = 8
    '''时区偏移量（相较于UTC+0时区）'''

    session_expire_time: int = 300
    '''会话过期时间（单位：秒）'''


get_setting, set_setting = make_setting_singleton(BaseSetting())



"""BlueFirmament Middleware"""

__all__ = [
    "BaseMiddleware",
]

import abc
import typing
from ..utils.main import call_as_async

if typing.TYPE_CHECKING:
    from blue_firmament.task.context import BaseTaskContext


type NextT = typing.Callable[[], typing.Coroutine[typing.Any, typing.Any, None]]
type MiddlewaresT = typing.List['BaseMiddleware']


class BaseMiddleware(abc.ABC):
    """Base class of BlueFirmament Middleware.
    """

    @abc.abstractmethod
    def __call__(self, *, next_: NextT, task_context: 'BaseTaskContext') -> typing.Union[
        None, typing.Coroutine
    ]:
        ...

    @staticmethod
    def run_middlewares(middlewares: MiddlewaresT, task_context: "BaseTaskContext"):
        return call_as_async(
            middlewares[0], 
            next_=BaseMiddleware._get_next(middlewares, task_context=task_context), 
            task_context=task_context
        )

    @staticmethod
    def _get_next(
        middlewares: MiddlewaresT, task_context: "BaseTaskContext",
        current: int = 0
    ) -> NextT:
        async def _next() -> None:
            nonlocal current
            current += 1
            if current < len(middlewares):
                return await call_as_async(
                    middlewares[current],
                    next=BaseMiddleware._get_next(
                        middlewares, task_context=task_context, current=current
                    ),
                    task_context=task_context
                )
            else:
                return None
        return _next




__all__ = [
    'SchemeMetaclass',
    'BaseScheme', 
    'SchemeTV',
    'NoProxyScheme',
    'BaseRootScheme',
    "merge"
]

import abc
import copy
import inspect
import types
import typing
from typing import Optional as Opt

from ..utils.typing_ import safe_issubclass
from .._types import Undefined, _undefined
from .validator import SchemeValidator, FieldValidator
from .field import (
    CompositeField, PrivateField, Field, 
    field, dump_field_name
)
from .field import FieldValueProxy
if typing.TYPE_CHECKING:
    from ..dal.types import DALPath
    from ..dal.query_components.filters import EqFilter
    from ..dal.query_components import DALQueryComponent
    from ..dal.base import DataAccessLayer
    from ..log import LoggerT


@typing.dataclass_transform(
    kw_only_default=True,
    field_specifiers=(
        Field, CompositeField,
        field
    )
)
class SchemeMetaclass(abc.ABCMeta):

    """碧霄数据模型元类

    Design Doc: :doc:`/design/scheme/index`

    Rational
    ---------
    - 允许数据模型通过类属性定义字段

    Attributes
    -----------
    - ``__table_name__``：表名
    - ``__schema_name__``：数据库名
    - ``__fields__``：字段字典

    Usage
    -------
    定义字段
    ^^^^^^^^^

    基本示例：

    ```python
    class MyScheme(BlueFirmamentScheme):
        table_name = 'my_table'
        
        _id: BlueFirmamentSequenceField = BlueFirmamentField(is_primary_key=True)  # 类型为自增的字段
        name: str = 'default_name'
    ```
    
    可以直接使用`BlueFirmamentField`定义字段，也可以直接使用值定义字段。值将会作为字段的默认值，类变量名将自动作为字段名。

    名称为下列的类变量会被解析为内置字段：（不会继承）
    - ``_table_name``：表名
    - ``_schema_name``：数据库名/表组名（来源于PostgreSQL，相当于MySQL的数据库）

    声明私有字段使用``BlueFirmamentPrivateField``

    数据模型实例化
    ^^^^^^^^^^^^^^^^
    - 未传递的字段将使用默认值，没有默认值将变为BlueFirmamentUndefinedValue

    Behavior
    ---------
    字段继承
    ^^^^^^^^^^
    - 普通字段和私有字段都会被继承
    - 按照MRO顺序继承（往后的优先级更高）
    - 如果声明了同名字段且不是一个字段实例，会复制并覆盖（默认值、名称）

    """

    __builtin_cvars__: typing.Dict[str, typing.Any] = {
        '__dal__': None,
        '__dal_path__': None, 
        '__proxy__': False,
        '__disable_log__': False,
        '__key__': None,
        '__fields__': dict,
        '__partial__': False,
        '__inherit_validators__': True,
        '__private_fields__': dict,
        '__scheme_validators__': list,
        '__after_field_validators__': list,
        '__default_edflags__': None,
        '__default_idflags__': None
    }
    """Builtin class variables of BlueFirmamentScheme
    
    values are the default value (callable for mutable values).
    """
    __builtin_ivars__: typing.Dict[str, typing.Any] = {
        '__logger__': None,
        '__instantiated__': False,
        '__unset_fields__': set,
        '__dirty_fields__': set,
        '__field_values__': dict,
    }
    """Builtin instance variables of BlueFirmamentScheme
    
    values are the default value (callable for mutable values).
    """

    def __new__(
        cls, name: str, 
        bases: typing.Tuple[type[typing.Any], ...], 
        attrs: typing.Dict[str, typing.Any],
        dal_path: Opt["DALPath"] = None,
        dal: Opt[typing.Type["DataAccessLayer"]] = None,
        proxy: Opt[bool] = None,
        disable_log: Opt[bool] = None,
        partial: Opt[bool] = None,
        inherit_validators: Opt[bool] = None,
        default_exclude_dump_flags: Opt[set[str]] = None,
        default_include_dump_flags: Opt[set[str]] = None,
        **kwargs
    ):

        # Exclude base scheme class
        if name in ("BaseScheme",):
            return super().__new__(cls, name, bases, attrs, **kwargs)

        # Set up class vars
        if dal_path:
            attrs["__dal_path__"] = dal_path
        if dal:
            attrs["__dal__"] = dal
        if proxy:
            attrs["__proxy__"] = proxy
        if disable_log:
            attrs["__disable_log__"] = disable_log
        if partial:
            attrs["__partial__"] = partial
        if inherit_validators:
            attrs["__inherit_validators__"] = inherit_validators
        if default_exclude_dump_flags:
            attrs["__default_edflags__"] = default_exclude_dump_flags
        if default_include_dump_flags:
            attrs["__default_idflags__"] = default_include_dump_flags
        for cvar, default_v in cls.__builtin_cvars__.items():
            if cvar not in attrs:
                # Find in bases
                found = False
                for base in reversed(bases):
                    if hasattr(base, cvar):
                        # FIXME multiple scheme to inherit from
                        if callable(default_v):
                            # mutable value
                            val = attrs.setdefault(cvar, default_v())
                            if isinstance(val, dict):
                                val.update(getattr(base, cvar))
                            elif isinstance(val, list):
                                val.extend(getattr(base, cvar))
                            else:
                                raise TypeError(
                                    f"Unsupported type {type(val)} for class variable {cvar}"
                                )
                        else:
                            # immutable value
                            attrs[cvar] = getattr(base, cvar)
                        
                        found = True
                        continue
                if not found:
                    attrs[cvar] = default_v if not callable(default_v) else default_v()
        
        # Resolve fields
        fields: typing.Dict[str, Field] = attrs['__fields__']
        fields.update({
            k: v.fork(fork_validators=attrs['__inherit_validators__'])
            for k, v in fields.items()
        })
        private_fields: typing.Dict[str, PrivateField] = attrs['__private_fields__']
        private_fields.update({
            k: v.fork(fork_validators=attrs['__inherit_validators__'])
            for k, v in private_fields.items()
        })
        scheme_validators: typing.List[SchemeValidator]
        if not attrs['__inherit_validators__']:
            scheme_validators = list()
            attrs['__scheme_validators__'] = scheme_validators
        else:
            scheme_validators = attrs['__scheme_validators__']

        # Resolve attrs
        for k, v in attrs.items():
            # Skip dunder methods
            if k.startswith('__') and k.endswith('__'):
                continue

            # Skip method and function
            if inspect.ismethoddescriptor(v):
                continue
            elif inspect.isfunction(v):
                continue

            # Skip property
            if isinstance(v, property):
                continue

            # Skip field validator(s)
            if isinstance(v, FieldValidator) or (
                isinstance(v, list) and isinstance(v[0], FieldValidator)
            ):
                continue

            # Resolve scheme validators
            if isinstance(v, SchemeValidator):
                scheme_validators.append(v)
                continue

            # Resolve private fields 
            if isinstance(v, PrivateField):
                private_fields[k] = v
                v._set_name(k, True) # 如果没有配置名称，则使用类变量名作为字段名
                v._set_in_scheme_name(k, True)
                continue

            # Resolve fields
            if isinstance(v, Field):
                # already a field instance
                fields[k] = v
                v._set_name(k, True) # TODO rename as try
                v._set_in_scheme_name(k, True)
            else:
                # not a field instance
                if k in fields:
                    fields[k] = fields[k].fork(
                        default=v, name=k, in_scheme_name=k
                    )
                    continue
                elif k in private_fields:
                    private_fields[k] = private_fields[k].fork(
                        default=v, name=k, in_scheme_name=k,
                    )
                    continue

                fields[k] = Field(v, name=k, in_scheme_name=k)

        # Set up field converter from annotations
        for k, field_ in (fields | private_fields).items():
            anno = attrs.get('__annotations__', {}).get(k)
            if anno:
                field_._set_converter_from_anno(anno)

        # Resolve attrs having only annotation
        cls_annotations = attrs.get('__annotations__', {})
        for k, anno in cls_annotations.items():

            if k.startswith('__') and k.endswith('__'):
                continue
            
            if k not in fields and k not in private_fields:
                # Field[ValueT]
                orig = typing.get_origin(anno)
                if safe_issubclass(orig, Field):
                    field = orig(
                        name=k, in_scheme_name=k, 
                        vtype=typing.get_args(anno)[0]
                    )
                    field._set_converter_from_anno(anno)
                    fields[k] = field
                    continue

                # ValueT
                # not yet resolved above
                fields[k] = Field(name=k, in_scheme_name=k, vtype=anno)
                fields[k]._set_converter_from_anno(anno)


        # Replace attributes that recognized as fields' value to field instance
        for k, default_v in (fields | private_fields).items():
            attrs[k] = default_v

        # Resolve key field
        for v in fields.values():
            if v.is_key(): 
                attrs['__key__'] = v
                break


        # dynamically create __init__ method
        init_params: set[str] = set()
        init_assignments = []
        new_globals = globals().copy()
        for k, field_ins in (fields | private_fields).items():
            # skip init=False
            if not field_ins.init:
                continue
            
            # type_str = cls._get_type_from_anno(cls_annotations, k, new_globals)
            init_params.add(k)
            if isinstance(field_ins, CompositeField):
                sub_scheme_name = field_ins._sub.__name__
                if field_ins.is_partial or attrs['__partial__']:
                    partial_sub_scheme = copy.copy(field_ins._sub)
                    setattr(partial_sub_scheme, '__partial__', True)
                    new_globals[sub_scheme_name] = partial_sub_scheme
                else:
                    new_globals[sub_scheme_name] = field_ins._sub

                for sub_field in field_ins.sub_fields:
                    init_params.add(sub_field.in_scheme_name)

                init_assignments.append(f"    self.{k} = {k} if {k} is not _undefined \
                    else {sub_scheme_name}({
                    ",".join(
                        f"{i.in_scheme_name}={i.in_scheme_name}"
                        for i in field_ins.sub_fields
                    )
                })")
            else:
                init_assignments.append(f"    self.{k} = {k}")
        
        if init_params:
            init_sig = f"def __init__(self, *, {','.join(f'{i}=_undefined' for i in init_params)}, **kwargs):\n"
        else:
            init_sig = "def __init__(self, **kwargs):\n"
        init_body = '\n'
        init_body += '    SchemeMetaclass.init_ivars(self)\n'
        init_body += '\n'.join(init_assignments)
        init_body += '\n    SchemeMetaclass.run_scheme_validators(self)\n'
        init_body += '    self.__instantiated__ = True\n'
        init_body += '    SchemeMetaclass.run_after_field_validators(self)\n'
        init_body += '    if not self.__disable_log__:\n'
        init_body += '        self._logger.info("Scheme instantiated", scheme_data=self.dump_to_dict())\n'
        init_body += '    self.__post_init__()\n'

        init_method = init_sig + init_body
        
        exec(init_method, new_globals, attrs)

        result_class = super().__new__(cls, name, bases, attrs, **kwargs)

        # set fields' scheme
        for k, default_v in (result_class.__fields__).items():
            default_v._set_scheme_cls(
                typing.cast(typing.Type["BaseScheme"], result_class), 
                no_raise=True, force=True
            )

        return result_class
    
    @staticmethod
    def _get_type_from_anno(cls_annotations, k: str, globols):
        type_: typing.Type | types.UnionType | None = cls_annotations.get(k, None)
        if type_ is None:
            type_str = 'typing.Any'
        elif isinstance(type_, types.UnionType):
            type_str = '|'.join([t.__name__ for t in type_.__args__])
            type_str = type_str.replace('NoneType', 'None')
        else:
            type_str = type_.__name__
            globols[type_str] = type_

        return type_str
    
    @staticmethod
    def init_ivars(ins: "BaseScheme"):
        """Initialize instance variables
        """
        for k, default_v in SchemeMetaclass.__builtin_ivars__.items():
            setattr(ins, k, default_v() if callable(default_v) else default_v)

    @staticmethod
    def run_scheme_validators(ins: "BaseScheme"):
        for validator in ins.__scheme_validators__:
            validator(ins)

    @staticmethod
    def run_after_field_validators(ins: "BaseScheme"):
        for _ in range(len(ins.__after_field_validators__)):
            validator = ins.__after_field_validators__.pop(0)
            validator(value=ins[validator._field], scheme_ins=ins)


TV = typing.TypeVar("TV")


class BaseScheme(metaclass=SchemeMetaclass):
    """Base data model class of BlueFirmament

    Features
    ---------
    Serialization
    ^^^^^^^^^^^^^

    Partial
    ^^^^^^^
    Partial field will remain undefined if not provided during instantiation and
    no default value set, otherwise raise `ValueError`.

    Set `__partial__` to `True` to make all fields partial, you can override this
    in each field definition.
    """

    # class vars
    __dal__: typing.ClassVar[Opt[typing.Type["DataAccessLayer"]]]
    __dal_path__: typing.ClassVar[Opt["DALPath"]]
    __key__: typing.ClassVar[Opt[Field]]  # TODO use field
    """Field that uniquely identifies an instance.

    Includes primary key, composite key.
    """
    __disable_log__: typing.ClassVar[bool]
    """Disable scheme internal logs
    """
    __proxy__: typing.ClassVar[bool]
    """Proxy field value or not

    If disabled:
    - Mutable field value modification will not be tracked
    - Cannot get scheme or field info from field value
    """
    __fields__: typing.ClassVar[typing.Dict[str, Field]]
    """Fields and their instances defined in the scheme
    """
    __private_fields__: typing.ClassVar[typing.Dict[str, PrivateField]]
    """Private fields and their instances defined in the scheme
    """
    __scheme_validators__: typing.ClassVar[typing.List[SchemeValidator]]
    __after_field_validators__: typing.ClassVar[typing.List[FieldValidator]]
    """Field validators needed to be ran
    immediately after scheme instantiation
    """
    __partial__: typing.ClassVar[bool]
    """If True, all fields are partial
    """
    __inherit_validators__: typing.ClassVar[bool]
    """If False, scheme and field validators will not be inherited by
    sub scheme.
    """
    __default_edflags__: typing.ClassVar[Opt[set[str]]]
    __default_idflags__: typing.ClassVar[Opt[set[str]]]

    # instance vars
    __field_values__: typing.ClassVar[typing.Dict[str, typing.Any]]
    """Storing each field's value
    
    - Key is the field's in_scheme_name.
    """
    __dirty_fields__: typing.ClassVar[typing.Set[str]]
    """Which fields are modified since last dump

    - Exclude private fields.
    """
    __unset_fields__: typing.ClassVar[typing.Set[str]]
    """Fields that are not provided during instantiation

    - Exclude private fields
    - Instance variable
    """
    __logger__: typing.ClassVar[Opt["LoggerT"]]
    """Scheme level logger

    - Instance variable
    """
    __instantiated__: typing.ClassVar[bool]
    """Whether the scheme is instantiated

    - Instance variable
    """

    def __post_init__(self) -> None:
        """数据模型实例化后执行的操作；可以被重写"""

    @staticmethod
    def _init_private_fields(obj: 'BaseScheme', data: typing.Any):

        for k, v in obj.__private_fields__.items():
            if k in data:
                setattr(obj, k, v.convert(data[k]))
            else:
                setattr(obj, k, v.default_value)

    @classmethod
    def __from_parents__(cls, /, *parents: "BaseScheme") -> typing.Self:
        """Instantiate this scheme from its parent schemes.

        Precondition:
        - Inherited fields' name unchanged.
        """
        data = {}
        for parent in parents:
            data.update(parent.dump_to_dict())

        return cls(**data)

    def _mark_unset(
        self, field_: str | Field
    ) -> None:
        """Mark field as unset (not provided during instantiation)
        """
        if isinstance(field_, Field):
            self.__unset_fields__.add(field_.in_scheme_name)
        else:
            self.__unset_fields__.add(field_)

    def _mark_dirty(self, field_: str | Field) -> None:
        """Mark field as dirty (modified since last dump)
        """
        if isinstance(field_, str):
            field_ = self.__fields__[field_]
        self.__dirty_fields__.add(field_.name)

    @classmethod
    def dal_path(cls) -> "DALPath":
        """
        :raise ValueError: when not set
        """
        if not cls.__dal_path__:
            raise ValueError("dal path not set")
        return cls.__dal_path__
    
    def equals(self) -> typing.Tuple["DALQueryComponent", ...]:
        """Get EqFilter of all fields.
        """
        return tuple(
            field.equals(self._get_value(field))
            for field in self.__fields__.values()
        )

    @classmethod
    def _get_key_field(cls) -> Field:
        """
        :raise KeyError: if no key on scheme
        """
        if not cls.__key__:
            raise KeyError(f'{cls.__name__} does not have a key')
        return cls.__key__

    @classmethod
    def _try_get_key_field(cls) -> Opt[Field]:
        """Try to get key field, return None if not exists.
        """
        return cls.__key__ if cls.__key__ else None
    
    @property
    def key_value(self) -> typing.Any:
        return self._get_value(self._get_key_field())
    
    @property
    def key_eqf(self) -> "EqFilter":
        key_field = self._get_key_field()
        return key_field.equals(self._get_value(key_field))
    
    # def dump(self, target_type: typing.Type[TV]) -> TV:
    #     """Serialize
    #     """
    #     if target_type is str:
    #         return self.dump_to_str()
    #     elif target_type is dict:
    #         return self.dump_to_dict()

    def __str__(self):
        return self.dump_to_str()

    def dump_to_str(
        self,
        use_name: bool = False
    ) -> str:
        """Serialize model to string

        :param use_name: use field's name instead of in_scheme_name,
                        defaults to False
        :type use_name: bool, optional
        """
        return ",".join(
            f"{in_name if not use_name else i.name}={\
                i.dump_val_to_str(FieldValueProxy.dump(self[i]))}"
            for in_name, i in self.__fields__.items()
        )

    def dump_to_dict(
        self,
        only_dirty: bool = False,
        exclude_natural_key: bool = False,
        exclude_unset: Opt[bool] = None,
        exclude_flags: Opt[set[str]] = None,
        include_flags: Opt[set[str]] = None,
        only_private: bool = False,
        jsonable: bool = True
    ) -> dict:
        """Serialize to (jsonable) dict

        :param only_dirty: 
            If True, dirty fields will be reset.
        :param exclude_natural_key:
            If True, exclude natural key field.
        :param exclude_unset:
            If True, exclude unset fields.
            If None and is partial scheme, defaults to True.
        :param exclude_flags:
            If provided, exclude fields with all these flags.
            If not provided, ``default_exclude_dump_flags`` configured on
            scheme will be used.
            Prior to ``include_flags`` (same for scheme default).
        :param include_flags:
            If provided, only dumps fields with all these flags.
            If not provided, ``default_include_dump_flags`` configured on
            scheme will be used.
        :param only_private:
            If True, only private fields will de dumped.
        :param jsonable:
            If True, ensure the return is jsonable.

        Behaviour
        ----------
        - 调用每个字段的校验器来序列化字段值
        """
        data = dict()
        field_names: typing.Set[str]

        if only_private:
            field_names = set(self.__private_fields__.keys())
            for k in field_names:
                data[k] = getattr(self, k)
        else:
            if only_dirty:
                field_names = self.__dirty_fields__
            else:
                field_names = set(self.__fields__.keys())

            if exclude_natural_key:
                key_field = self._try_get_key_field()
                if key_field and key_field.is_key_natural():
                    field_names = field_names - {key_field.in_scheme_name}

            if exclude_unset is True or (exclude_unset is None and self.__partial__):
                field_names = field_names - self.__unset_fields__

            if exclude_flags is None and self.__default_edflags__:
                exclude_flags = self.__default_edflags__

            if include_flags is None and self.__default_idflags__:
                include_flags = self.__default_idflags__

            for k in field_names:
                field: Field = self.__fields__[k]

                if exclude_flags:
                    if field.dump_flags.issuperset(exclude_flags):
                        continue

                if include_flags and not exclude_flags:
                    if not field.dump_flags.issuperset(include_flags):
                        continue

                field_v = FieldValueProxy.dump(getattr(self, k))
                if jsonable:
                    if isinstance(field, CompositeField):
                        data.update(field.dump_val_to_jsonable(field_v))
                    else:
                        data[k] = field.dump_val_to_jsonable(field_v)
                else:
                    data[k] = field_v

        return data
    
    def __getitem__(self, key: str | Field) -> typing.Any:
        """通过字段名/字段获取字段值

        Note: 不可以是其他属性，只可以是字段
        """
        if key in self.__fields__:
            return getattr(self, dump_field_name(key, in_scheme=True))
        
        raise KeyError(f'{key} is not a field of {self.__class__.__name__}')
    
    def __setitem__(self, key: str | Field, value: typing.Any) -> None:
        """通过字段/字段名设置字段值

        Note: 不可以是其他属性，只可以是字段
        """
        field = self.__fields__[dump_field_name(key)]
        field.__set__(self, value)

    @classmethod
    def keys(cls) -> typing.Iterable[str]:
        return cls.__fields__.keys()
    
    def values(self) -> typing.Iterable[typing.Any]:
        return self.__field_values__.values()
    
    FieldValueType = typing.TypeVar("FieldValueType")

    def _set_value(
        self, 
        field: Field[FieldValueType], 
        value: "FieldValueProxy[FieldValueType]" | FieldValueType | Undefined
    ) -> None:
        """设置字段值

        :param field: 字段名或字段实例
        """
        self.__field_values__[field.in_scheme_name] = value

    def _get_value(
        self,
        field: Field[FieldValueType]
    ) -> "FieldValueProxy[FieldValueType]" | FieldValueType:
        """获取字段值

        :param field: 字段名或字段实例
        """
        return self.__field_values__[field.in_scheme_name]

    def _merge(self, scheme: "BaseScheme") -> None:
        """Merge current scheme with another scheme's values.

        Use ``dump_to_dict`` to get the values of the other scheme.

        :param scheme: The scheme to update from
        """
        if not isinstance(scheme, BaseScheme):
            raise TypeError(f"Expected BaseScheme, got {type(scheme)}")

        dumped = scheme.dump_to_dict()
        for field_name, value in dumped.items():
            self.__fields__[field_name].__set__(self, value)

    @property
    def _logger(self):
        """Scheme level logger

        Context:
        - scheme_id: id(self)
        """
        if not self.__logger__:
            from ..log import get_logger
            logger = get_logger(self.__class__.__name__)
            logger = logger.bind(
                scheme_id=id(self),
            )
            self.__logger__ = logger
        return self.__logger__
    
    def _set_logger(self, logger: "LoggerT") -> None:
        """Set scheme level logger"""
        self.__logger__ = logger


SchemeTV = typing.TypeVar('SchemeTV', bound=BaseScheme)
"""数据模型类型变量
"""


class NoProxyScheme(BaseScheme,
    proxy=False                    
):
    pass


class BaseRootScheme(BaseScheme):

    """根数据模型类

    - 只有一个 root 字段
    - 序列化时 root 不会作为字段名，而是直接序列化 root 字段的值
    """

    root: Field


def merge(scheme1: BaseScheme, scheme2: BaseScheme) -> None:
    """Merge same fields (by name)'s value from scheme2 to scheme1.

    Will firstly dump scheme2 (so partial will not be included)

    Examples
    --------
    >>> merge(SchemeA(a=1, b=3), SchemeB(a=2))
    SchemeA: a=2, b=3
    >>> merge(SchemeA(a=1), SchemeB(a=_undefined))
    SchemeB: a=1
    """
    for field_ in scheme2.dump_to_dict():
        try:
            scheme1[field_] = scheme2[field_]
        except KeyError:
            continue




import abc
from typing import Optional as Opt

from blue_firmament.dal import DALPath


class DALQueryComponent(abc.ABC):

    @abc.abstractmethod
    def dump_to_sql(self) -> str:
        """序列化为SQL语句
        """

    def dump_to_postgrest(self) -> tuple[
        str, tuple | dict | None
    ]:
        """Serialize to a call to postgrest lib's RequestBuilder.

        :returns: (RequestBuilder method, parameters)

        Parameters：
        - tuple as positional arguments
        - dict as keyword arguments
        - None for no parameters
        """

    @abc.abstractmethod
    def dump_to_postgrest_str(self) -> str:
        """Serialize to PostgREST query part in string.

        Syntax: `<field>.<filter_name>.<value>`,
        - use `()` for in filters with multiple values,
        - use `{}` for contains(By) filters with multiple values.

        Filter names:
        - eq: equal
        - neq: not equal
        - is: is (true, false, null)
        - in: in (multiple values)
        - cs: contains (multiple values)
        - cb: contained by (multiple values)
        """

    def __repr__(self):
        return f"{self.__class__.__name__}"

    @property
    def dal_path(self) -> Opt[DALPath]:
        """Get filter's scheme's DAL path.
        """
        return None



"""请求上下文
"""

__all__ = [
    "BaseTaskContext", 
    "ExtendedTaskContext",
    "SoBaseTC",
]

import contextvars
import typing
from typing import Optional as Opt
from ...session import SessionTV
from ...scheme.main import SchemeTV
from ...scheme import BaseScheme, private_field, FieldT

if typing.TYPE_CHECKING:
    from ...log import LoggerT
    from .. import Task
    from ..result import TaskResult


class BaseTaskContextFields(typing.TypedDict):
    task: typing.NotRequired["Task"]
    task_result: typing.NotRequired["TaskResult"]
    base_logger: typing.NotRequired["LoggerT"]
    """Bind task context based on this logger.
    """


class BaseTaskContext:
    """Base class of TaskContext.

    .. versionchanged:: 0.1.2
        rename to BaseTaskContext from RequestContext and
        removed fields from session.
    """

    def __init__(
        self,
        btc: Opt["BaseTaskContext"] = None,
        **kwargs: typing.Unpack[BaseTaskContextFields]
    ):
        """
        :param btc: another BaseTaskContext instance to copy from
        """
        if btc:
            self.__task = btc._task
            self.__task_result = btc._task_result
            self.__logger = btc._logger
        else:
            self.__task = kwargs["task"]
            self.__task_result = kwargs["task_result"]
            self.__logger = kwargs["base_logger"].bind(
                trace_id=self.__task.trace_id,
            )

    @property
    def _task(self) -> "Task": return self.__task
    @property
    def _task_result(self) -> 'TaskResult': return self.__task_result
    @property
    def _logger(self) -> "LoggerT": return self.__logger
    @_logger.setter
    def _logger(self, new_logger: "LoggerT"): self.__logger = new_logger
    
    CONTEXTVAR = contextvars.ContextVar[typing.Self]('TASKC_CONTEXTVAR')
    @classmethod
    def set_contextvar(cls, task_context: typing.Self) -> None:
        cls.CONTEXTVAR.set(task_context)
    @classmethod
    def from_contextvar(cls) -> typing.Self:
        """
        :raise LookupError: if not set
        """
        return cls.CONTEXTVAR.get()


class ExtendedTaskContext(
    typing.Generic[SessionTV],
    BaseTaskContext, 
):
    """Extend BaseTaskContext with session.
    """

    def __init_subclass__(
        cls,
        session_cls: Opt[typing.Type[SessionTV]] = None
    ) -> None:
        if session_cls:
            cls.__session_cls = session_cls
        super().__init_subclass__()

    def __init__(
        self,
        tc: BaseTaskContext | typing.Self,
        skip_btc_init: bool = False
    ):
        """
        :param tc: BaseTaskContext or ExtendedTaskContext instance.
        :param skip_btc_init: For manager's sake!
        """
        if not skip_btc_init:
            super().__init__(btc=tc)
        if tc.__class__ is BaseTaskContext:
            self.__session: SessionTV = self.__session_cls.from_task(tc._task)
        elif isinstance(tc, ExtendedTaskContext):
            self.__session: SessionTV = tc._session
        else:
            raise TypeError("tc must be either BaseTaskContext or subclass of ExtendedTaskContext")
        self.__init_fields__()

    def __init_fields__(self):
        """Assign your customized fields.

        Better to use property instead of set attribute.
        """

    @property
    def _session(self): 
        return self.__session


class SoBaseTC(BaseScheme):
    """Scheme attached BaseTaskContext.

    By inheriting this class, your class can access
    task context and its properties with ease.

    Or by inheriting this class then override ``_task_context``'s
    type to your customized TaskContext and add fields, enables
    your scheme accessing your customized task context.

    .. versionchanged:: 0.1.2
        rename to ``SoBTC`` from ``SchemeHasRequestContext``
    """

    _task_context: FieldT[BaseTaskContext] = private_field(
        default_factory=BaseTaskContext.from_contextvar
    )

    def __post_init__(self) -> None:
        # update scheme logger context
        self._set_logger(self._logger.bind(
            **self._task_context._logger._context
        ))

    @property
    def _task(self):
        return self._task_context._task
    @property
    def _task_result(self):
        return self._task_context._task_result



import datetime
import typing

from ...utils.datetime_ import get_datetimez
from ...scheme import private_field
from ...setting import EnvJsonSetting, make_setting_singleton

type AuthSessionField = typing.Literal[
    "session_id",
    "user_id",
    "roles",
    "expire_at",
    "issued_at"
]
SESSION_FIELDS_GETTER_DEFAULT: dict[AuthSessionField, typing.Callable[[dict], typing.Any]] = {
    "session_id": lambda payload: payload["sid"],
    "expire_at": lambda payload: get_datetimez(timestamp=payload["exp"]),
    "issued_at": lambda payload: get_datetimez(timestamp=payload["iat"]),
    "user_id": lambda payload: payload["sub"],
    "roles": lambda payload: payload["roles"],
}

class AuthSetting(EnvJsonSetting):

    _setting_name = private_field(default="auth")
    _is_packaged = private_field(default=False)

    jwt_secret_key: str = ""
    jwt_algorithms: tuple[str, ...] = ("HS256",)
    jwt_allowed_audiences: tuple[str, ...] = ()
    session_fields_getter: dict[AuthSessionField, typing.Callable[[dict], typing.Any]] = \
        SESSION_FIELDS_GETTER_DEFAULT


get_setting, set_setting = make_setting_singleton(AuthSetting.load())



"""DAL Utils"""


import typing
from .types import QueryComLikeType, FieldLikeType
from ..utils.enum_ import dump_enum

if typing.TYPE_CHECKING:
    from .query_components import DALQueryComponent


def dump_query_coms_like(
    *values: QueryComLikeType,
    scheme_like: typing.Any | None = None
) -> typing.Iterable["DALQueryComponent"]:

    """Convert list of FilterLikeType to DALFilter list

    :param scheme_like: BaseScheme 类, BlueFirmamentField 实例

    Behavior
    --------
    - 如果 item of value 是字符串或整数，会被转换为 EqFilter
        - 使用 scheme 的主键作为字段键
    """
    res = []

    from ..scheme import BaseScheme
    from ..scheme.field import Field
    for item in values:
        if item is None:
            continue
        if isinstance(item, (str, int)):
            try:
                if scheme_like:
                    if isinstance(scheme_like, Field):
                        scheme_like = scheme_like.scheme_cls
                    if issubclass(scheme_like, BaseScheme) or isinstance(scheme_like, BaseScheme):
                        res.append(scheme_like._get_key_field().equals(item))
                        continue

                raise ValueError
            except (AttributeError, ValueError):
                raise ValueError(
                    "Cannot dump filter-like value that is not a DALQueryComponent without scheme"
                )
        elif isinstance(item, BaseScheme):
            res.extend(item.equals())
        else:
            res.append(item)

    return typing.cast(typing.Iterable["DALQueryComponent"], res)


def dump_field_like(value: FieldLikeType) -> str:
    """Convert FieldLikeType to field name string

    :param value: to dump value
    :param dal_path: optional, Path to execute current query.
        If provided, will prepend the field name with `scheme_dal_path[0].` if
        the `dal_path` is different from the scheme's dal_path.
        (for resolving reference)
    """
    from ..scheme.field import dump_field_name
    return dump_field_name(dump_enum(value))


"""Transporter listening to a Queue.
"""
import typing
from .base import BaseTransporter
from ..task import Task, TaskResult

if typing.TYPE_CHECKING:
    from ..dal.base import QueueLikeDataAccessLayer
    from .. import BlueFirmamentApp


class QueueTransporter(BaseTransporter):
    """

    :ivar __dal: Listening this queue dal for tasks.
    :ivar __handling_dal: The queue storing handling tasks.
    """

    def __init__(
        self,
        app: BlueFirmamentApp,
        queue_dal: QueueLikeDataAccessLayer,
        handling_queue_dal: QueueLikeDataAccessLayer,
        name: str = "default"
    ):
        super().__init__(app=app, name=name)
        self.__dal = queue_dal
        self.__handling_dal = handling_queue_dal
        self.__stop = False

    async def start(self):
        while not self.__stop:
            await self(await self.__dal.pop())

    async def stop(self):
        self.__stop = True

    async def __call__(self, raw: bytes):
        await self._app.handle_task(
            task=Task.load_from_bytes(raw),
            task_result=TaskResult(),
            transporter=self
        )
